import pytest
from unittest.mock import patch

from hera.vm import VirtualMachine


@pytest.fixture
def vm():
    return VirtualMachine()


def test_SWI_emits_warning(vm):
    with patch("hera.utils._emit_msg") as mock_emit_warning:
        vm.exec_swi(1)
        assert mock_emit_warning.call_count == 1
        assert "Warning" in mock_emit_warning.call_args[0][0]
        assert "SWI is a no-op in this simulator" in mock_emit_warning.call_args[0][0]


def test_SWI_increments_pc(vm):
    vm.exec_swi(1)
    assert vm.pc == 1


def test_RTI_emits_warning(vm):
    with patch("hera.utils._emit_msg") as mock_emit_warning:
        vm.exec_rti()
        assert "Warning" in mock_emit_warning.call_args[0][0]
        assert "RTI is a no-op in this simulator" in mock_emit_warning.call_args[0][0]


def test_RTI_increments_pc(vm):
    vm.exec_rti()
    assert vm.pc == 1



import pytest
from unittest.mock import patch

from hera.data import Op
from hera.vm import VirtualMachine


@pytest.fixture
def vm():
    return VirtualMachine()


def test_exec_one_delegates_to_savef(vm):
    with patch("hera.vm.VirtualMachine.exec_savef") as mock_exec_savef:
        vm.exec_one(Op("SAVEF", ["R1"]))
        assert mock_exec_savef.call_count == 1
        assert mock_exec_savef.call_args == (("R1",), {})


def test_savef_with_sign(vm):
    vm.flag_sign = True
    vm.exec_savef("R5")
    assert vm.registers[5] == 1
    assert vm.flag_sign


def test_savef_with_zero(vm):
    vm.flag_zero = True
    vm.exec_savef("R5")
    assert vm.registers[5] == 0b10
    assert vm.flag_zero


def test_savef_with_overflow(vm):
    vm.flag_overflow = True
    vm.exec_savef("R5")
    assert vm.registers[5] == 0b100
    assert vm.flag_overflow


def test_savef_with_carry(vm):
    vm.flag_carry = True
    vm.exec_savef("R5")
    assert vm.registers[5] == 0b1000
    assert vm.flag_carry


def test_savef_with_carry_block(vm):
    vm.flag_carry_block = True
    vm.exec_savef("R5")
    assert vm.registers[5] == 0b10000


def test_savef_with_several_flags(vm):
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.exec_savef("R5")
    assert vm.registers[5] == 0b1101
    assert vm.flag_sign
    assert vm.flag_overflow
    assert vm.flag_carry


def test_savef_with_all_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_savef("R5")
    assert vm.registers[5] == 0b11111
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_savef_with_no_flags(vm):
    vm.exec_savef("R5")
    assert vm.registers[5] == 0


def test_savef_overwrites_high_bits(vm):
    vm.registers[5] = 17500
    vm.exec_savef("R5")
    assert vm.registers[5] == 0


def test_savef_increments_pc(vm):
    vm.exec_savef("R5")
    assert vm.pc == 1


def test_savef_does_not_affect_R0(vm):
    vm.flag_carry = True
    vm.exec_savef("R0")
    assert vm.registers[0] == 0


def test_exec_one_delegates_to_rstrf(vm):
    with patch("hera.vm.VirtualMachine.exec_rstrf") as mock_exec_rstrf:
        vm.exec_one(Op("RSTRF", ["R1"]))
        assert mock_exec_rstrf.call_count == 1
        assert mock_exec_rstrf.call_args == (("R1",), {})


def test_rstrf_with_sign(vm):
    vm.registers[5] = 1
    vm.exec_rstrf("R5")
    assert vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_rstrf_with_zero(vm):
    vm.registers[5] = 0b10
    vm.exec_rstrf("R5")
    assert not vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_rstrf_with_overflow(vm):
    vm.registers[5] = 0b100
    vm.exec_rstrf("R5")
    assert not vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_rstrf_with_carry(vm):
    vm.registers[5] = 0b1000
    vm.exec_rstrf("R5")
    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block


def test_rstrf_with_carry_block(vm):
    vm.registers[5] = 0b10000
    vm.exec_rstrf("R5")
    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert vm.flag_carry_block


def test_rstrf_with_several_flags(vm):
    vm.registers[5] = 0b1101
    vm.exec_rstrf("R5")
    assert vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block


def test_rstrf_with_all_flags(vm):
    vm.registers[5] = 0b11111
    vm.exec_rstrf("R5")
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_rstrf_with_no_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_rstrf("R5")
    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_rstrf_increments_pc(vm):
    vm.exec_rstrf("R5")
    assert vm.pc == 1


def test_exec_one_delegates_to_fon(vm):
    with patch("hera.vm.VirtualMachine.exec_fon") as mock_exec_fon:
        vm.exec_one(Op("FON", [5]))
        assert mock_exec_fon.call_count == 1
        assert mock_exec_fon.call_args == ((5,), {})


def test_fon_with_sign(vm):
    vm.exec_fon(1)
    assert vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_fon_with_zero(vm):
    vm.exec_fon(0b10)
    assert not vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_fon_with_overflow(vm):
    vm.exec_fon(0b100)
    assert not vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_fon_with_carry(vm):
    vm.exec_fon(0b1000)
    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block


def test_fon_with_carry_block(vm):
    vm.exec_fon(0b10000)
    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert vm.flag_carry_block


def test_fon_with_multiple_flags(vm):
    vm.exec_fon(0b10101)
    assert vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert not vm.flag_carry
    assert vm.flag_carry_block


def test_fon_with_no_flags(vm):
    vm.exec_fon(0)
    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_fon_does_not_overwrite_flags(vm):
    vm.flag_carry_block = True
    vm.exec_fon(1)
    assert vm.flag_sign
    assert vm.flag_carry_block


def test_fon_increments_pc(vm):
    vm.exec_fon(0)
    assert vm.pc == 1


def test_exec_one_delegates_to_foff(vm):
    with patch("hera.vm.VirtualMachine.exec_foff") as mock_exec_foff:
        vm.exec_one(Op("FOFF", [5]))
        assert mock_exec_foff.call_count == 1
        assert mock_exec_foff.call_args == ((5,), {})


def test_foff_with_sign(vm):
    vm.flag_zero = True
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_foff(1)
    assert not vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_foff_with_zero(vm):
    vm.flag_zero = True
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_foff(0b10)
    assert vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_foff_with_overflow(vm):
    vm.flag_zero = True
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_foff(0b100)
    assert vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_foff_with_carry(vm):
    vm.flag_zero = True
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_foff(0b1000)
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert not vm.flag_carry
    assert vm.flag_carry_block


def test_foff_with_carry_block(vm):
    vm.flag_zero = True
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_foff(0b10000)
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block


def test_foff_with_multiple_flags(vm):
    vm.flag_zero = True
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_foff(0b10101)
    assert not vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block


def test_foff_with_no_flags(vm):
    vm.flag_zero = True
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_foff(0)
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_foff_increments_pc(vm):
    vm.exec_foff(0)
    assert vm.pc == 1


def test_exec_one_delegates_to_fset5(vm):
    with patch("hera.vm.VirtualMachine.exec_fset5") as mock_exec_fset5:
        vm.exec_one(Op("FSET5", [5]))
        assert mock_exec_fset5.call_count == 1
        assert mock_exec_fset5.call_args == ((5,), {})


def test_fset5_with_sign(vm):
    vm.exec_fset5(1)
    assert vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_fset5_with_zero(vm):
    vm.exec_fset5(0b10)
    assert not vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_fset5_with_overflow(vm):
    vm.exec_fset5(0b100)
    assert not vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_fset5_with_carry(vm):
    vm.exec_fset5(0b1000)
    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block


def test_fset5_with_carry_block(vm):
    vm.exec_fset5(0b10000)
    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert vm.flag_carry_block


def test_fset5_with_multiple_flags(vm):
    vm.exec_fset5(0b10101)
    assert vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert not vm.flag_carry
    assert vm.flag_carry_block


def test_fset5_with_no_flags(vm):
    vm.exec_fset5(0)
    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_fset5_does_overwrite_flags(vm):
    vm.flag_zero = True
    vm.flag_carry_block = True
    vm.exec_fset5(1)
    assert vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_carry_block


def test_fset5_increments_pc(vm):
    vm.exec_fset5(0)
    assert vm.pc == 1


def test_exec_one_delegates_to_fset4(vm):
    with patch("hera.vm.VirtualMachine.exec_fset4") as mock_exec_fset4:
        vm.exec_one(Op("FSET4", [5]))
        assert mock_exec_fset4.call_count == 1
        assert mock_exec_fset4.call_args == ((5,), {})


def test_fset4_with_sign(vm):
    vm.exec_fset4(1)
    assert vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_fset4_with_zero(vm):
    vm.exec_fset4(0b10)
    assert not vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_fset4_with_overflow(vm):
    vm.exec_fset4(0b100)
    assert not vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_fset4_with_carry(vm):
    vm.exec_fset4(0b1000)
    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block


def test_fset4_with_multiple_flags(vm):
    vm.exec_fset4(0b101)
    assert vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_fset4_with_no_flags(vm):
    vm.exec_fset4(0)
    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_fset4_does_overwrite_flags(vm):
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.exec_fset4(1)
    assert vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow


def test_fset4_increments_pc(vm):
    vm.exec_fset4(0)
    assert vm.pc == 1



"""Parse HERA programs.

`parse` and its wrapper `parse_file` are the two functions intended for public use.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: December 2018
"""
import os
import re
import sys

from lark import Lark, Token as LarkToken, Transformer, Tree
from lark.exceptions import LarkError, UnexpectedCharacters, UnexpectedToken

from .data import IntToken, Location, Op, Token
from .utils import emit_error, emit_warning, get_canonical_path, is_register


def parse(text, *, fpath=None, expand_includes=True, visited=None):
    """Parse a HERA program from a string into a list of Op objects. If errors are
    encountered while parsing, an error message is emitted and the whole program exits.

    `fpath` is the path of the file being parsed, as it will appear in error and
    debugging messages. It defaults to "<string>".

    `expand_includes` determines what happens when an #include statement is encountered.
    If `expand_includes` is True, then the #include statement is interpreted as it is by
    the C preprocessor, i.e. the file identified by #include's argument is read, parsed,
    and pasted in with the rest of the operations. If it is False, then the #include
    statement is retained as an Op object.

    `visited` is a set of file paths that have already been visited, to prevent infinite
    regress if two files include each other directly or indirectly.
    """
    if visited is None:
        visited = set()

    if fpath is not None:
        visited.add(get_canonical_path(fpath))

    file_lines = text.splitlines()
    base_location = Location(None, None, fpath or "<string>", file_lines)

    try:
        tree = _parser.parse(text)
    except UnexpectedCharacters as e:
        loc = base_location._replace(line=e.line, column=e.column)
        emit_error("unexpected character", loc=loc, exit=True)
    except UnexpectedToken as e:
        if e.token.type == "$END":
            emit_error("unexpected end of file", exit=True)
        else:
            loc = base_location._replace(line=e.line, column=e.column)
            emit_error("unexpected character", loc=loc, exit=True)
    except LarkError as e:
        loc = base_location._replace(line=e.line, column=e.column)
        emit_error("invalid syntax", loc=base_location, exit=True)

    if isinstance(tree, Tree):
        ops = tree.children
    elif isinstance(tree, Op):
        ops = [tree]
    else:
        ops = tree

    convert_tokens(ops, base_location)

    if expand_includes:
        expanded_ops = []
        for op in ops:
            if op.name == "#include" and len(op.args) == 1:
                if op.args[0].startswith("<"):
                    # TODO: Probably need to handle these somehow.
                    continue

                # Strip off the leading and trailing quote.
                include_path = op.args[0][1:-1]
                include_path = os.path.join(os.path.dirname(fpath), include_path)

                if get_canonical_path(include_path) in visited:
                    # TODO: Do I _need_ to exit immediately here, or can I catch more
                    # errors?
                    emit_error("recursive include", loc=op.args[0].location, exit=True)

                expanded_ops.extend(parse_file(include_path, visited=visited))
            else:
                expanded_ops.append(op)
        ops = expanded_ops

    return ops


def parse_file(fpath, *, expand_includes=True, allow_stdin=False, visited=None):
    """Convenience function for parsing a HERA file. Reads the contents of the file and
    delegates parsing to the `parse` function.

    `allow_stdin` should be set to True if you wish the file path "-" to be interpreted
    as standard input instead of a file with that actual name. See `parse` for the
    meaning of `expand_includes` and `visited`.
    """
    if allow_stdin and fpath == "-":
        try:
            program = sys.stdin.read()
        except (IOError, KeyboardInterrupt):
            print()
            sys.exit(3)
    else:
        try:
            with open(fpath) as f:
                program = f.read()
        except FileNotFoundError:
            emit_error('file "{}" does not exist.'.format(fpath), exit=True)
        except PermissionError:
            emit_error('permission denied to open file "{}".'.format(fpath), exit=True)
        except OSError:
            emit_error('could not open file "{}".'.format(fpath), exit=True)

    return parse(program, fpath=fpath, expand_includes=expand_includes, visited=visited)


def convert_tokens(ops, base_location):
    """Convert all tokens in the list of ops from Lark Token objects to HERA Token and
    IntToken objects, tagged with `base_location`.
    """
    for i, op in enumerate(ops):
        name = Token(op.name.type, op.name, augment_location(base_location, op.name))
        for j, arg in enumerate(op.args):
            if arg.type == "DECIMAL" or arg.type == "CHAR":
                op.args[j] = IntToken(arg, augment_location(base_location, arg))
            elif arg.type == "HEX":
                op.args[j] = IntToken(
                    arg, augment_location(base_location, arg), base=16
                )
            elif arg.type == "OCTAL":
                if not arg.startswith("0o"):
                    msg = "zero-prefixed numbers are interpreted as octal"
                    msg += " (consider using 0o prefix instead)"
                    loc = augment_location(base_location, arg)
                    emit_warning(msg, loc=loc),
                op.args[j] = IntToken(arg, augment_location(base_location, arg), base=8)
            elif arg.type == "BINARY":
                op.args[j] = IntToken(arg, augment_location(base_location, arg), base=2)
            else:
                op.args[j] = Token(arg.type, arg, augment_location(base_location, arg))
        ops[i] = op._replace(name=name)


class TreeToOplist(Transformer):
    """Transform Lark's parse tree into a list of HERA ops."""

    def start(self, matches):
        # TODO: Figure out why all this is necessary.
        if len(matches) == 2:
            if isinstance(matches[0], tuple):
                return [matches[0]] + matches[1]
            else:
                return matches[0] + matches[1]
        elif len(matches) > 2:
            return matches[:-1] + matches[-1]
        else:
            return matches[0]

    def cpp_program(self, matches):
        emit_warning("void HERA_main() { ... } is not necessary")
        return matches

    def hera_program(self, matches):
        return matches

    def op(self, matches):
        return Op(matches[0], matches[1:])

    def include(self, matches):
        return Op(LarkToken("SYMBOL", "#include"), [matches[0]])

    def value(self, matches):
        line = matches[0].line
        column = matches[0].column

        if matches[0].type == "STRING":
            otkn = matches[0]
            s = replace_escapes(otkn[1:-1])
            ntkn = LarkToken("STRING", s)
            # Preserve data from the original token.
            ntkn.pos_in_stream = otkn.pos_in_stream
            ntkn.line = otkn.line
            ntkn.column = otkn.column
            ntkn.end_line = otkn.end_line
            ntkn.end_column = otkn.end_column
            return ntkn
        elif matches[0].type == "CHAR":
            # Strip the leading and trailing quote.
            s = matches[0][1:-1]
            if s.startswith("\\"):
                c = char_to_escape(s[1])
            else:
                c = s
            return LarkToken("CHAR", ord(c), line=line, column=column)
        elif matches[0].type == "SYMBOL":
            if is_register(matches[0]):
                matches[0].type = "REGISTER"
            return matches[0]
        else:
            return matches[0]


_parser = Lark(
    r"""
    start: include* (cpp_program | hera_program)

    cpp_program: _cpp_open (op | include)* _CPP_CLOSE
    hera_program: op (op | include)* |

    // Underscores before everything so that no tokens end up in the tree.
    _cpp_open: "void" _SYMBOL "(" ")" "{"
    _CPP_CLOSE: "}"
    _SYMBOL: SYMBOL

    include: "#include" ( STRING | /<[^>]+>/ )

    op: SYMBOL "(" _arglist? ")" ";"?

    _arglist: ( value "," )* value

    value: DECIMAL | HEX | OCTAL | BINARY | SYMBOL | STRING | CHAR

    SYMBOL: /[A-Za-z_][A-Za-z0-9_]*/
    DECIMAL: /-?[1-9][0-9]*/ | "0"
    HEX.2: /-?0x[0-9a-fA-F]+/
    OCTAL.2: /-?0o[0-7]+/ | /-?0[1-9]+/
    BINARY.2: /-?0b[01]+/
    STRING: /"(\\.|[^"])*"/
    CHAR: /'(\\.|.)'/

    COMMENT: ( "//" /[^\n]*/ | "/*" /([^*]|\*[^\/])*/ "*/" )

    %import common.WS
    %ignore WS
    %ignore COMMENT
    """,
    parser="lalr",
    transformer=TreeToOplist(),
)


def augment_location(base_location, token):
    """Add the line and column information from the token to the Location object."""
    return base_location._replace(line=token.line, column=token.column)


def replace_escapes(s):
    """Replace all backslash escape sequences in the string with the special characters
    that they encode.
    """
    return re.sub(r"\\.", lambda m: char_to_escape(m.group(0)[1]), s)


def char_to_escape(c):
    """Return the special character that `c` encodes.

        >>> char_to_escape("n")
        "\n"
    """
    if c == "n":
        return "\n"
    elif c == "t":
        return "\t"
    elif c == "\\":
        return "\\"
    elif c == '"':
        return '"'
    else:
        # TODO: Give a warning for this.
        return "\\" + c



from hera.main import main
from hera.symtab import HERA_DATA_START
from hera.vm import VirtualMachine


def test_addition_program(capsys):
    vm = VirtualMachine()
    main(["test/assets/simple/addition.hera"], vm)

    assert vm.registers[1] == 20
    assert vm.registers[2] == 22
    assert vm.registers[3] == 42

    for r in vm.registers[4:]:
        assert r == 0

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block

    for x in vm.memory:
        assert x == 0

    assert "Warning" not in capsys.readouterr().err


def test_loop_program(capsys):
    vm = VirtualMachine()
    main(["test/assets/simple/loop.hera"], vm)

    assert vm.registers[1] == 10
    assert vm.registers[2] == 10

    for r in vm.registers[3:10]:
        assert r == 0

    assert not vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block

    for x in vm.memory:
        assert x == 0

    assert "Warning" not in capsys.readouterr().err


def test_function_call_program(capsys):
    vm = VirtualMachine()
    main(["test/assets/simple/function_call.hera"], vm)

    assert vm.registers[1] == 16

    for r in vm.registers[2:10]:
        assert r == 0

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block

    for x in vm.memory:
        assert x == 0

    assert "Warning" not in capsys.readouterr().err


def test_fibonacci_program(capsys):
    vm = VirtualMachine()
    main(["test/assets/simple/fibonacci.hera"], vm)

    assert vm.registers[1] == 12
    assert vm.registers[2] == 144
    assert vm.registers[3] == 89
    assert vm.registers[4] == 12
    assert vm.registers[5] == 89

    for r in vm.registers[6:10]:
        assert r == 0

    assert not vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block

    for x in vm.memory:
        assert x == 0

    assert "Warning" not in capsys.readouterr().err


def test_data_easy_program(capsys):
    vm = VirtualMachine()
    main(["test/assets/simple/data_easy.hera"], vm)

    assert vm.registers[1] == HERA_DATA_START
    assert vm.registers[2] == 42

    for r in vm.registers[3:]:
        assert r == 0

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block

    assert vm.memory[HERA_DATA_START] == 42

    assert "Warning" not in capsys.readouterr().err


def test_dskip_program(capsys):
    vm = VirtualMachine()
    main(["test/assets/simple/dskip.hera"], vm)

    assert vm.registers[1] == HERA_DATA_START
    assert vm.registers[2] == 42
    assert vm.registers[3] == 84

    for r in vm.registers[4:]:
        assert r == 0

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block

    assert vm.memory[HERA_DATA_START] == 42
    assert vm.memory[HERA_DATA_START + 11] == 84

    assert "Warning" not in capsys.readouterr().err


def test_loop_and_constant_program(capsys):
    vm = VirtualMachine()
    main(["test/assets/simple/loop_and_constant.hera"], vm)

    assert vm.registers[1] == 100
    assert vm.registers[2] == 100
    assert vm.registers[3] == 5050

    for r in vm.registers[4:10]:
        assert r == 0

    assert not vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block

    assert "Warning" not in capsys.readouterr().err


def test_hera_boilerplate_program():
    vm = VirtualMachine()
    main(["test/assets/simple/hera_boilerplate.hera"], vm)

    assert vm.registers[1] == 42
    assert vm.registers[2] == 42

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block



import pytest

from hera.vm import VirtualMachine


@pytest.fixture
def vm():
    return VirtualMachine()


def test_print_reg_increments_pc(vm):
    vm.exec_print_reg("R1")
    assert vm.pc == 1


def test_print_increments_pc(vm):
    vm.exec_print("Hello, world!")
    assert vm.pc == 1


def test_println_increments_pc(vm):
    vm.exec_println("Hello, world!")
    assert vm.pc == 1



"""Type-check HERA programs.

`typecheck` is the public interface of this module.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: December 2018
"""
from .data import Token
from .utils import (
    DATA_STATEMENTS,
    emit_error,
    is_symbol,
    register_to_index,
    RELATIVE_BRANCHES,
)


def typecheck(program, symtab):
    """Type-check the program and emit errors as appropriate."""
    current_file = None
    end_of_data = False
    for op in program:
        # Reset the end_of_data flag whenever an op from a new file is encountered.
        if isinstance(op.name, Token) and op.name.location is not None:
            if current_file is None:
                current_file = op.name.location.path
            else:
                if current_file != op.name.location.path:
                    end_of_data = False
                    current_file = op.name.location.path

        if not end_of_data:
            if op.name not in DATA_STATEMENTS:
                end_of_data = True
        else:
            if op.name in DATA_STATEMENTS:
                emit_error("data statement after instruction", loc=op.name.location)

        typecheck_one(op, symtab)

        if op.name in RELATIVE_BRANCHES:
            if len(op.args) == 1 and is_symbol(op.args[0]):
                msg = "relative branches cannot use labels"
                msg += " (why not use {} instead?)".format(op.name[:-1])
                emit_error(msg, loc=op.args[0].location)


def typecheck_one(op, symtab):
    """Type-check a single HERA operation and emit errors as appropriate."""
    params = _types_map.get(op.name)
    if params is not None:
        check_types(op.name, params, op.args, symtab)
    else:
        emit_error("unknown instruction `{}`".format(op.name), loc=op.name.location)


def check_types(name, expected, got, symtab):
    """Verify that the given args match the expected ones and emit errors as
    appropriate. `name` is the name of the HERA op, as a Token object. `expected` is a
    tuple or list of constants (REGISTER, U16, etc., defined above) representing the
    expected argument types to the operation. `args` is a tuple or list of the actual
    arguments given.
    """
    if len(got) < len(expected):
        emit_error(
            "too few args to {} (expected {})".format(name, len(expected)),
            loc=name.location,
        )

    if len(expected) < len(got):
        emit_error(
            "too many args to {} (expected {})".format(name, len(expected)),
            loc=name.location,
        )

    ordinals = ["first", "second", "third"]
    for ordinal, pattern, arg in zip(ordinals, expected, got):
        prefix = "{} arg to {} ".format(ordinal, name)
        error = check_one_type(pattern, arg, symtab)
        if error:
            emit_error(prefix + error, loc=arg.location)


def check_one_type(pattern, arg, symtab):
    """Verify that the argument matches the pattern. Return a string stating the error
    if it doesn't, return None otherwise.
    """
    # TODO: Overengineered?
    if pattern == REGISTER:
        if not isinstance(arg, Token) or arg.type != "REGISTER":
            return "not a register"

        if arg.lower() == "pc":
            return "program counter cannot be accessed or changed directly"

        try:
            register_to_index(arg)
        except ValueError:
            return "not a valid register"
    elif pattern == REGISTER_OR_LABEL:
        if not isinstance(arg, Token):
            return "not a register or label"

        if arg.type == "REGISTER":
            if arg.lower() == "pc":
                return "program counter cannot be accessed or changed directly"

            try:
                register_to_index(arg)
            except ValueError:
                return "not a valid register"
        elif arg.type != "SYMBOL":
            return "not a register or label"
    elif pattern == LABEL:
        if not isinstance(arg, Token) or arg.type != "SYMBOL":
            return "not a symbol"
    elif pattern == STRING:
        if not isinstance(arg, Token) or arg.type != "STRING":
            return "not a string"
    elif isinstance(pattern, range):
        if is_symbol(arg):
            try:
                arg = symtab[arg]
            except KeyError:
                return "undefined constant"

        if not isinstance(arg, int):
            return "not an integer"
        if arg not in pattern:
            if pattern.start == 0 and arg < 0:
                return "must not be negative"
            else:
                return "out of range"
    else:
        raise RuntimeError(
            "unknown pattern in hera.typechecker.check_one_type", pattern
        )


# Constants to pass to check_types
REGISTER = "r"
REGISTER_OR_LABEL = "rl"
LABEL = "l"
STRING = "s"
U4 = range(0, 2 ** 4)
U5 = range(0, 2 ** 5)
U16 = range(0, 2 ** 16)
I8 = range(-2 ** 7, 2 ** 8)
I16 = range(-2 ** 15, 2 ** 16)


_types_map = {
    # Set and increment instructions
    "SETLO": (REGISTER, I8),
    "SETHI": (REGISTER, I8),
    "INC": (REGISTER, range(1, 65)),
    "DEC": (REGISTER, range(1, 65)),
    # Arithmetic, logical and shift instructions
    "AND": (REGISTER, REGISTER, REGISTER),
    "OR": (REGISTER, REGISTER, REGISTER),
    "ADD": (REGISTER, REGISTER, REGISTER),
    "SUB": (REGISTER, REGISTER, REGISTER),
    "MUL": (REGISTER, REGISTER, REGISTER),
    "XOR": (REGISTER, REGISTER, REGISTER),
    "LSL": (REGISTER, REGISTER),
    "LSR": (REGISTER, REGISTER),
    "LSL8": (REGISTER, REGISTER),
    "LSR8": (REGISTER, REGISTER),
    "ASL": (REGISTER, REGISTER),
    "ASR": (REGISTER, REGISTER),
    # Flag manipulation
    "SAVEF": (REGISTER,),
    "RSTRF": (REGISTER,),
    "FON": (U5,),
    "FOFF": (U5,),
    "FSET5": (U5,),
    "FSET4": (U4,),
    # Memory instructions
    "LOAD": (REGISTER, U5, REGISTER),
    "STORE": (REGISTER, U5, REGISTER),
    # Branch instructions
    "CALL": (REGISTER, REGISTER_OR_LABEL),
    "RETURN": (REGISTER, REGISTER_OR_LABEL),
    "BR": (REGISTER_OR_LABEL,),
    "BRR": (I8,),
    "BL": (REGISTER_OR_LABEL,),
    "BLR": (I8,),
    "BGE": (REGISTER_OR_LABEL,),
    "BGER": (I8,),
    "BLE": (REGISTER_OR_LABEL,),
    "BLER": (I8,),
    "BG": (REGISTER_OR_LABEL,),
    "BGR": (I8,),
    "BULE": (REGISTER_OR_LABEL,),
    "BULER": (I8,),
    "BUG": (REGISTER_OR_LABEL,),
    "BUGR": (I8,),
    "BZ": (REGISTER_OR_LABEL,),
    "BZR": (I8,),
    "BNZ": (REGISTER_OR_LABEL,),
    "BNZR": (I8,),
    "BC": (REGISTER_OR_LABEL,),
    "BCR": (I8,),
    "BNC": (REGISTER_OR_LABEL,),
    "BNCR": (I8,),
    "BS": (REGISTER_OR_LABEL,),
    "BSR": (I8,),
    "BNS": (REGISTER_OR_LABEL,),
    "BNSR": (I8,),
    "BV": (REGISTER_OR_LABEL,),
    "BVR": (I8,),
    "BNV": (REGISTER_OR_LABEL,),
    "BNVR": (I8,),
    # Interrupt processing
    "SWI": (U4,),
    "RTI": (),
    # Pseudo-instructions
    "SET": (REGISTER, I16),
    "SETRF": (REGISTER, I16),
    "MOVE": (REGISTER, REGISTER),
    "CMP": (REGISTER, REGISTER),
    "NEG": (REGISTER, REGISTER),
    "NOT": (REGISTER, REGISTER),
    "CBON": (),
    "CON": (),
    "COFF": (),
    "CCBOFF": (),
    "FLAGS": (REGISTER,),
    "NOP": (),
    "HALT": (),
    "LABEL": (LABEL,),
    # Data statements
    "CONSTANT": (LABEL, I16),
    "DLABEL": (LABEL,),
    "INTEGER": (I16,),
    "LP_STRING": (STRING,),
    "DSKIP": (U16,),
    # Debugging instructions
    "print_reg": (REGISTER,),
    "print": (STRING,),
    "println": (STRING,),
}



"""Some important data structures for the HERA interpreter.

Location objects represent a location in a file, for the use of warnings and error
messages. The `file_lines` field is a pointer to a list of lines in the file, so that
the actual line can be printed in the message.

Op objects represent HERA operations.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: December 2018
"""
from collections import namedtuple


Location = namedtuple("Location", ["line", "column", "path", "file_lines"])


class Op(namedtuple("Op", ["name", "args", "original"])):
    def __new__(cls, name, args, original=None):
        return tuple.__new__(cls, (name, args, original))

    def __eq__(self, other):
        return (
            isinstance(other, Op)
            and self.name == other.name
            and self.args == other.args
        )

    def __repr__(self):
        orepr = "None" if self.original is None else "Op(...)"
        return "Op(name={!r}, args={!r}, original={})".format(
            self.name, self.args, orepr
        )


class IntToken(int):
    def __new__(cls, value, loc=None, **kwargs):
        self = super(IntToken, cls).__new__(cls, value, **kwargs)
        self.location = loc
        return self


class Token(str):
    def __new__(cls, type_, value, loc=None):
        self = super(Token, cls).__new__(cls, value)
        self.type = type_
        self.location = loc
        return self



import pytest
from unittest.mock import patch

from hera.data import Op
from hera.vm import VirtualMachine


@pytest.fixture
def vm():
    return VirtualMachine()


def test_exec_one_delegates_to_br(vm):
    with patch("hera.vm.VirtualMachine.exec_br") as mock_exec_br:
        vm.exec_one(Op("BR", ["R1"]))
        assert mock_exec_br.call_count == 1
        assert mock_exec_br.call_args == (("R1",), {})


def test_br_sets_pc(vm):
    vm.registers[7] = 170
    vm.exec_br("R7")
    assert vm.pc == 170


def test_br_sets_pc_to_zero(vm):
    vm.exec_br("R0")
    assert vm.pc == 0


def test_br_does_not_change_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.registers[7] = 92
    vm.exec_br("R7")
    assert vm.pc == 92
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_brr(vm):
    with patch("hera.vm.VirtualMachine.exec_brr") as mock_exec_brr:
        vm.exec_one(Op("BRR", [12]))
        assert mock_exec_brr.call_count == 1
        assert mock_exec_brr.call_args == ((12,), {})


def test_brr_sets_pc(vm):
    vm.exec_brr(100)
    assert vm.pc == 100


def test_brr_with_negative_offset(vm):
    vm.pc = 50
    vm.exec_brr(-17)
    assert vm.pc == 33


def test_brr_sets_pc_with_previous_value(vm):
    vm.pc = 100
    vm.exec_brr(15)
    assert vm.pc == 115


def test_brr_sets_pc_to_zero(vm):
    vm.exec_brr(0)
    assert vm.pc == 0


def test_brr_does_not_change_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_brr(16)
    assert vm.pc == 16
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bl(vm):
    with patch("hera.vm.VirtualMachine.exec_bl") as mock_exec_bl:
        vm.exec_one(Op("BL", ["R3"]))
        assert mock_exec_bl.call_count == 1
        assert mock_exec_bl.call_args == (("R3",), {})


def test_bl_branches_on_sign(vm):
    vm.flag_sign = True
    vm.registers[3] = 47
    vm.exec_bl("R3")
    assert vm.pc == 47


def test_bl_branches_on_overflow(vm):
    vm.flag_overflow = True
    vm.registers[3] = 47
    vm.exec_bl("R3")
    assert vm.pc == 47


def test_bl_does_not_branch_on_sign_and_overflow(vm):
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.registers[3] = 47
    vm.exec_bl("R3")
    assert vm.pc == 1


def test_bl_does_not_branch_on_neither_sign_nor_overflow(vm):
    vm.registers[3] = 47
    vm.exec_bl("R3")
    assert vm.pc == 1


def test_bl_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bl("R0")
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_blr(vm):
    with patch("hera.vm.VirtualMachine.exec_blr") as mock_exec_blr:
        vm.exec_one(Op("BLR", [20]))
        assert mock_exec_blr.call_count == 1
        assert mock_exec_blr.call_args == ((20,), {})


def test_blr_branches_on_sign(vm):
    vm.flag_sign = True
    vm.pc = 100
    vm.exec_blr(20)
    assert vm.pc == 120


def test_blr_branches_on_overflow(vm):
    vm.flag_overflow = True
    vm.pc = 100
    vm.exec_blr(-20)
    assert vm.pc == 80


def test_blr_does_not_branch_on_sign_and_overflow(vm):
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.pc = 100
    vm.exec_blr(20)
    assert vm.pc == 101


def test_blr_does_not_branch_on_neither_sign_nor_overflow(vm):
    vm.pc = 100
    vm.exec_blr(20)
    assert vm.pc == 101


def test_blr_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_blr(20)
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bge(vm):
    with patch("hera.vm.VirtualMachine.exec_bge") as mock_exec_bge:
        vm.exec_one(Op("BGE", ["R3"]))
        assert mock_exec_bge.call_count == 1
        assert mock_exec_bge.call_args == (("R3",), {})


def test_bge_branches_on_no_flags(vm):
    vm.registers[3] = 47
    vm.exec_bge("R3")
    assert vm.pc == 47


def test_bge_branches_on_sign_and_overflow(vm):
    vm.flag_overflow = True
    vm.flag_sign = True
    vm.registers[3] = 47
    vm.exec_bge("R3")
    assert vm.pc == 47


def test_bge_does_not_branch_on_sign(vm):
    vm.flag_sign = True
    vm.registers[3] = 47
    vm.exec_bge("R3")
    assert vm.pc == 1


def test_bge_does_not_branch_on_overflow(vm):
    vm.flag_overflow = True
    vm.registers[3] = 47
    vm.exec_bge("R3")
    assert vm.pc == 1


def test_bge_branches_on_zero(vm):
    vm.flag_zero = True
    vm.registers[3] = 47
    vm.exec_bge("R3")
    assert vm.pc == 47


def test_bge_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bge("R0")
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bger(vm):
    with patch("hera.vm.VirtualMachine.exec_bger") as mock_exec_bger:
        vm.exec_one(Op("BGER", [20]))
        assert mock_exec_bger.call_count == 1
        assert mock_exec_bger.call_args == ((20,), {})


def test_bger_branches_on_no_flags(vm):
    vm.exec_bger(47)
    assert vm.pc == 47


def test_bger_branches_on_sign_and_overflow(vm):
    vm.flag_overflow = True
    vm.flag_sign = True
    vm.pc = 100
    vm.exec_bger(47)
    assert vm.pc == 147


def test_bger_does_not_branch_on_sign(vm):
    vm.flag_sign = True
    vm.exec_bger(47)
    assert vm.pc == 1


def test_bger_does_not_branch_on_overflow(vm):
    vm.flag_overflow = True
    vm.pc = 100
    vm.exec_bger(47)
    assert vm.pc == 101


def test_bger_branches_on_zero(vm):
    vm.flag_zero = True
    vm.exec_bger(47)
    assert vm.pc == 47


def test_bger_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bger(47)
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_ble(vm):
    with patch("hera.vm.VirtualMachine.exec_ble") as mock_exec_ble:
        vm.exec_one(Op("BLE", ["R3"]))
        assert mock_exec_ble.call_count == 1
        assert mock_exec_ble.call_args == (("R3",), {})


def test_ble_branches_on_sign(vm):
    vm.flag_sign = True
    vm.registers[3] = 47
    vm.exec_ble("R3")
    assert vm.pc == 47


def test_ble_branches_on_overflow(vm):
    vm.flag_overflow = True
    vm.registers[3] = 47
    vm.exec_ble("R3")
    assert vm.pc == 47


def test_ble_branches_on_zero(vm):
    vm.flag_zero = True
    vm.registers[3] = 47
    vm.exec_ble("R3")
    assert vm.pc == 47


def test_ble_branches_on_overflow_and_zero(vm):
    vm.flag_overflow = True
    vm.flag_zero = True
    vm.registers[3] = 47
    vm.exec_ble("R3")
    assert vm.pc == 47


def test_ble_does_not_branch_on_no_flags(vm):
    vm.registers[3] = 47
    vm.exec_ble("R3")
    assert vm.pc == 1


def test_ble_does_not_branch_on_sign_and_overflow(vm):
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.registers[3] = 47
    vm.exec_ble("R3")
    assert vm.pc == 1


def test_ble_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_ble("R0")
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bler(vm):
    with patch("hera.vm.VirtualMachine.exec_bler") as mock_exec_bler:
        vm.exec_one(Op("BLER", [47]))
        assert mock_exec_bler.call_count == 1
        assert mock_exec_bler.call_args == ((47,), {})


def test_bler_branches_on_sign(vm):
    vm.flag_sign = True
    vm.exec_bler(47)
    assert vm.pc == 47


def test_bler_branches_on_overflow(vm):
    vm.flag_overflow = True
    vm.pc = 100
    vm.exec_bler(47)
    assert vm.pc == 147


def test_bler_branches_on_zero(vm):
    vm.flag_zero = True
    vm.exec_bler(47)
    assert vm.pc == 47


def test_bler_branches_on_overflow_and_zero(vm):
    vm.flag_overflow = True
    vm.flag_zero = True
    vm.exec_bler(47)
    assert vm.pc == 47


def test_bler_does_not_branch_on_no_flags(vm):
    vm.exec_bler(47)
    assert vm.pc == 1


def test_bler_does_not_branch_on_sign_and_overflow(vm):
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.pc = 100
    vm.exec_bler(47)
    assert vm.pc == 101


def test_bler_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bler(47)
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bg(vm):
    with patch("hera.vm.VirtualMachine.exec_bg") as mock_exec_bg:
        vm.exec_one(Op("BG", ["R3"]))
        assert mock_exec_bg.call_count == 1
        assert mock_exec_bg.call_args == (("R3",), {})


def test_bg_branches_on_no_flags(vm):
    vm.registers[3] = 47
    vm.exec_bg("R3")
    assert vm.pc == 47


def test_bg_branches_on_sign_and_overflow(vm):
    vm.flag_overflow = True
    vm.flag_sign = True
    vm.registers[3] = 47
    vm.exec_bg("R3")
    assert vm.pc == 47


def test_bg_does_not_branch_on_sign(vm):
    vm.flag_sign = True
    vm.registers[3] = 47
    vm.exec_bg("R3")
    assert vm.pc == 1


def test_bg_does_not_branch_on_overflow(vm):
    vm.flag_overflow = True
    vm.registers[3] = 47
    vm.exec_bg("R3")
    assert vm.pc == 1


def test_bg_does_not_branch_on_zero(vm):
    vm.flag_zero = True
    vm.registers[3] = 47
    vm.exec_bg("R3")
    assert vm.pc == 1


def test_bg_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bg("R0")
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bgr(vm):
    with patch("hera.vm.VirtualMachine.exec_bgr") as mock_exec_bgr:
        vm.exec_one(Op("BGR", [20]))
        assert mock_exec_bgr.call_count == 1
        assert mock_exec_bgr.call_args == ((20,), {})


def test_bgr_branches_on_no_flags(vm):
    vm.exec_bgr(47)
    assert vm.pc == 47


def test_bgr_branches_on_sign_and_overflow(vm):
    vm.flag_overflow = True
    vm.flag_sign = True
    vm.pc = 100
    vm.exec_bgr(47)
    assert vm.pc == 147


def test_bgr_does_not_branch_on_sign(vm):
    vm.flag_sign = True
    vm.exec_bgr(47)
    assert vm.pc == 1


def test_bgr_does_not_branch_on_overflow(vm):
    vm.flag_overflow = True
    vm.pc = 100
    vm.exec_bgr(47)
    assert vm.pc == 101


def test_bgr_does_not_branch_on_zero(vm):
    vm.flag_zero = True
    vm.exec_bgr(47)
    assert vm.pc == 1


def test_bgr_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bgr(47)
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bule(vm):
    with patch("hera.vm.VirtualMachine.exec_bule") as mock_exec_bule:
        vm.exec_one(Op("BULE", ["R3"]))
        assert mock_exec_bule.call_count == 1
        assert mock_exec_bule.call_args == (("R3",), {})


def test_bule_branches_on_not_carry(vm):
    vm.flag_carry = False
    vm.registers[3] = 47
    vm.exec_bule("R3")
    assert vm.pc == 47


def test_bule_does_not_branch_on_overflow(vm):
    vm.flag_carry = True
    vm.flag_overflow = True
    vm.registers[3] = 47
    vm.exec_bule("R3")
    assert vm.pc == 1


def test_bule_branches_on_zero(vm):
    vm.flag_carry = True
    vm.flag_zero = True
    vm.registers[3] = 47
    vm.exec_bule("R3")
    assert vm.pc == 47


def test_bule_does_not_branch_on_sign(vm):
    vm.flag_carry = True
    vm.flag_sign = True
    vm.registers[3] = 47
    vm.exec_bule("R3")
    assert vm.pc == 1


def test_bule_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bule("R0")
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_buler(vm):
    with patch("hera.vm.VirtualMachine.exec_buler") as mock_exec_buler:
        vm.exec_one(Op("BULER", ["R3"]))
        assert mock_exec_buler.call_count == 1
        assert mock_exec_buler.call_args == (("R3",), {})


def test_buler_branches_on_not_carry(vm):
    vm.flag_carry = False
    vm.exec_buler(47)
    assert vm.pc == 47


def test_buler_does_not_branch_on_overflow(vm):
    vm.flag_carry = True
    vm.flag_overflow = True
    vm.pc = 100
    vm.exec_buler(47)
    assert vm.pc == 101


def test_buler_branches_on_zero(vm):
    vm.flag_carry = True
    vm.flag_zero = True
    vm.pc = 100
    vm.exec_buler(47)
    assert vm.pc == 147


def test_buler_does_not_branch_on_sign(vm):
    vm.flag_carry = True
    vm.flag_sign = True
    vm.exec_buler(47)
    assert vm.pc == 1


def test_buler_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_buler(47)
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bug(vm):
    with patch("hera.vm.VirtualMachine.exec_bug") as mock_exec_bug:
        vm.exec_one(Op("BUG", ["R3"]))
        assert mock_exec_bug.call_count == 1
        assert mock_exec_bug.call_args == (("R3",), {})


def test_bug_branches_on_carry_and_not_zero(vm):
    vm.flag_carry = True
    vm.flag_zero = False
    vm.registers[3] = 47
    vm.exec_bug("R3")
    assert vm.pc == 47


def test_bug_does_not_branch_on_carry_and_zero(vm):
    vm.flag_carry = True
    vm.flag_zero = True
    vm.registers[3] = 47
    vm.exec_bug("R3")
    assert vm.pc == 1


def test_bug_does_not_branch_on_not_carry_and_not_zero(vm):
    vm.registers[3] = 47
    vm.exec_bug("R3")
    assert vm.pc == 1


def test_bug_does_not_branch_on_sign(vm):
    vm.flag_sign = True
    vm.registers[3] = 47
    vm.exec_bug("R3")
    assert vm.pc == 1


def test_bug_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bug("R0")
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bugr(vm):
    with patch("hera.vm.VirtualMachine.exec_bugr") as mock_exec_bugr:
        vm.exec_one(Op("BUGR", [47]))
        assert mock_exec_bugr.call_count == 1
        assert mock_exec_bugr.call_args == ((47,), {})


def test_bugr_branches_on_carry_and_not_zero(vm):
    vm.flag_carry = True
    vm.flag_zero = False
    vm.exec_bugr(47)
    assert vm.pc == 47


def test_bugr_does_not_branch_on_carry_and_zero(vm):
    vm.flag_carry = True
    vm.flag_zero = True
    vm.pc = 100
    vm.exec_bugr(47)
    assert vm.pc == 101


def test_bugr_does_not_branch_on_not_carry_and_not_zero(vm):
    vm.exec_bugr(47)
    assert vm.pc == 1


def test_bugr_does_not_branch_on_sign(vm):
    vm.flag_sign = True
    vm.exec_bugr(47)
    assert vm.pc == 1


def test_bugr_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bugr(47)
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bz(vm):
    with patch("hera.vm.VirtualMachine.exec_bz") as mock_exec_bz:
        vm.exec_one(Op("BZ", ["R3"]))
        assert mock_exec_bz.call_count == 1
        assert mock_exec_bz.call_args == (("R3",), {})


def test_bz_branches_on_zero(vm):
    vm.flag_zero = True
    vm.registers[3] = 47
    vm.exec_bz("R3")
    assert vm.pc == 47


def test_bz_does_not_branch_on_not_zero(vm):
    vm.registers[3] = 47
    vm.exec_bz("R3")
    assert vm.pc == 1


def test_bz_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bz("R3")
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bzr(vm):
    with patch("hera.vm.VirtualMachine.exec_bzr") as mock_exec_bzr:
        vm.exec_one(Op("BZR", [47]))
        assert mock_exec_bzr.call_count == 1
        assert mock_exec_bzr.call_args == ((47,), {})


def test_bzr_branches_on_zero(vm):
    vm.flag_zero = True
    vm.pc = 100
    vm.exec_bzr(47)
    assert vm.pc == 147


def test_bzr_does_not_branch_on_not_zero(vm):
    vm.exec_bzr(47)
    assert vm.pc == 1


def test_bzr_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bzr(47)
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bnz(vm):
    with patch("hera.vm.VirtualMachine.exec_bnz") as mock_exec_bnz:
        vm.exec_one(Op("BNZ", ["R3"]))
        assert mock_exec_bnz.call_count == 1
        assert mock_exec_bnz.call_args == (("R3",), {})


def test_bnz_branches_on_not_zero(vm):
    vm.registers[3] = 47
    vm.exec_bnz("R3")
    assert vm.pc == 47


def test_bnz_does_not_branch_on_zero(vm):
    vm.flag_zero = True
    vm.registers[3] = 47
    vm.exec_bnz("R3")
    assert vm.pc == 1


def test_bnz_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bnz("R3")
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bnzr(vm):
    with patch("hera.vm.VirtualMachine.exec_bnzr") as mock_exec_bnzr:
        vm.exec_one(Op("BNZR", [47]))
        assert mock_exec_bnzr.call_count == 1
        assert mock_exec_bnzr.call_args == ((47,), {})


def test_bnzr_branches_on_not_zero(vm):
    vm.pc = 100
    vm.exec_bnzr(47)
    assert vm.pc == 147


def test_bnzr_does_not_branch_on_zero(vm):
    vm.flag_zero = True
    vm.exec_bnzr(47)
    assert vm.pc == 1


def test_bnzr_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bnzr(47)
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bc(vm):
    with patch("hera.vm.VirtualMachine.exec_bc") as mock_exec_bc:
        vm.exec_one(Op("BC", ["R3"]))
        assert mock_exec_bc.call_count == 1
        assert mock_exec_bc.call_args == (("R3",), {})


def test_bc_branches_on_carry(vm):
    vm.flag_carry = True
    vm.registers[3] = 47
    vm.exec_bc("R3")
    assert vm.pc == 47


def test_bc_does_not_branch_on_not_carry(vm):
    vm.registers[3] = 47
    vm.exec_bc("R3")
    assert vm.pc == 1


def test_bc_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bc("R3")
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bcr(vm):
    with patch("hera.vm.VirtualMachine.exec_bcr") as mock_exec_bcr:
        vm.exec_one(Op("BCR", [47]))
        assert mock_exec_bcr.call_count == 1
        assert mock_exec_bcr.call_args == ((47,), {})


def test_bcr_branches_on_carry(vm):
    vm.flag_carry = True
    vm.pc = 100
    vm.exec_bcr(47)
    assert vm.pc == 147


def test_bcr_does_not_branch_on_not_carry(vm):
    vm.exec_bcr(47)
    assert vm.pc == 1


def test_bcr_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bcr(47)
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bnc(vm):
    with patch("hera.vm.VirtualMachine.exec_bnc") as mock_exec_bnc:
        vm.exec_one(Op("BNC", ["R3"]))
        assert mock_exec_bnc.call_count == 1
        assert mock_exec_bnc.call_args == (("R3",), {})


def test_bnc_branches_on_not_carry(vm):
    vm.registers[3] = 47
    vm.exec_bnc("R3")
    assert vm.pc == 47


def test_bnc_does_not_branch_on_carry(vm):
    vm.flag_carry = True
    vm.registers[3] = 47
    vm.exec_bnc("R3")
    assert vm.pc == 1


def test_bnc_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bnc("R3")
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bncr(vm):
    with patch("hera.vm.VirtualMachine.exec_bncr") as mock_exec_bncr:
        vm.exec_one(Op("BNCR", [47]))
        assert mock_exec_bncr.call_count == 1
        assert mock_exec_bncr.call_args == ((47,), {})


def test_bncr_branches_on_not_carry(vm):
    vm.pc = 100
    vm.exec_bncr(47)
    assert vm.pc == 147


def test_bncr_does_not_branch_on_carry(vm):
    vm.flag_carry = True
    vm.exec_bncr(47)
    assert vm.pc == 1


def test_bncr_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bncr(47)
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bs(vm):
    with patch("hera.vm.VirtualMachine.exec_bs") as mock_exec_bs:
        vm.exec_one(Op("BS", ["R3"]))
        assert mock_exec_bs.call_count == 1
        assert mock_exec_bs.call_args == (("R3",), {})


def test_bs_branches_on_sign(vm):
    vm.flag_sign = True
    vm.registers[3] = 47
    vm.exec_bs("R3")
    assert vm.pc == 47


def test_bs_does_not_branch_on_not_sign(vm):
    vm.registers[3] = 47
    vm.exec_bs("R3")
    assert vm.pc == 1


def test_bs_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bs("R3")
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bsr(vm):
    with patch("hera.vm.VirtualMachine.exec_bsr") as mock_exec_bsr:
        vm.exec_one(Op("BSR", [47]))
        assert mock_exec_bsr.call_count == 1
        assert mock_exec_bsr.call_args == ((47,), {})


def test_bsr_branches_on_sign(vm):
    vm.flag_sign = True
    vm.pc = 100
    vm.exec_bsr(47)
    assert vm.pc == 147


def test_bsr_does_not_branch_on_not_sign(vm):
    vm.exec_bsr(47)
    assert vm.pc == 1


def test_bsr_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bsr(47)
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bns(vm):
    with patch("hera.vm.VirtualMachine.exec_bns") as mock_exec_bns:
        vm.exec_one(Op("BNS", ["R3"]))
        assert mock_exec_bns.call_count == 1
        assert mock_exec_bns.call_args == (("R3",), {})


def test_bns_branches_on_not_sign(vm):
    vm.registers[3] = 47
    vm.exec_bns("R3")
    assert vm.pc == 47


def test_bns_does_not_branch_on_sign(vm):
    vm.flag_sign = True
    vm.registers[3] = 47
    vm.exec_bns("R3")
    assert vm.pc == 1


def test_bns_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bns("R3")
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bnsr(vm):
    with patch("hera.vm.VirtualMachine.exec_bnsr") as mock_exec_bnsr:
        vm.exec_one(Op("BNSR", [47]))
        assert mock_exec_bnsr.call_count == 1
        assert mock_exec_bnsr.call_args == ((47,), {})


def test_bnsr_branches_on_not_sign(vm):
    vm.pc = 100
    vm.exec_bnsr(47)
    assert vm.pc == 147


def test_bnsr_does_not_branch_on_sign(vm):
    vm.flag_sign = True
    vm.exec_bnsr(47)
    assert vm.pc == 1


def test_bnsr_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bnsr(47)
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bv(vm):
    with patch("hera.vm.VirtualMachine.exec_bv") as mock_exec_bv:
        vm.exec_one(Op("BV", ["R3"]))
        assert mock_exec_bv.call_count == 1
        assert mock_exec_bv.call_args == (("R3",), {})


def test_bv_branches_on_overflow(vm):
    vm.flag_overflow = True
    vm.registers[3] = 47
    vm.exec_bv("R3")
    assert vm.pc == 47


def test_bv_does_not_branch_on_not_overflow(vm):
    vm.registers[3] = 47
    vm.exec_bv("R3")
    assert vm.pc == 1


def test_bv_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bv("R3")
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bvr(vm):
    with patch("hera.vm.VirtualMachine.exec_bvr") as mock_exec_bvr:
        vm.exec_one(Op("BVR", [47]))
        assert mock_exec_bvr.call_count == 1
        assert mock_exec_bvr.call_args == ((47,), {})


def test_bvr_branches_on_overflow(vm):
    vm.flag_overflow = True
    vm.pc = 100
    vm.exec_bvr(47)
    assert vm.pc == 147


def test_bvr_does_not_branch_on_not_overflow(vm):
    vm.exec_bvr(47)
    assert vm.pc == 1


def test_bvr_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bvr(47)
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bnv(vm):
    with patch("hera.vm.VirtualMachine.exec_bnv") as mock_exec_bnv:
        vm.exec_one(Op("BNV", ["R3"]))
        assert mock_exec_bnv.call_count == 1
        assert mock_exec_bnv.call_args == (("R3",), {})


def test_bnv_branches_on_not_overflow(vm):
    vm.registers[3] = 47
    vm.exec_bnv("R3")
    assert vm.pc == 47


def test_bnv_does_not_branch_on_overflow(vm):
    vm.flag_overflow = True
    vm.registers[3] = 47
    vm.exec_bnv("R3")
    assert vm.pc == 1


def test_bnv_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bnv("R3")
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_bnvr(vm):
    with patch("hera.vm.VirtualMachine.exec_bnvr") as mock_exec_bnvr:
        vm.exec_one(Op("BNVR", [47]))
        assert mock_exec_bnvr.call_count == 1
        assert mock_exec_bnvr.call_args == ((47,), {})


def test_bnvr_branches_on_not_overflow(vm):
    vm.pc = 100
    vm.exec_bnvr(47)
    assert vm.pc == 147


def test_bnvr_does_not_branch_on_overflow(vm):
    vm.flag_overflow = True
    vm.exec_bnvr(47)
    assert vm.pc == 1


def test_bnvr_does_not_set_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_bnvr(47)
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_exec_one_delegates_to_call(vm):
    with patch("hera.vm.VirtualMachine.exec_call") as mock_exec_call:
        vm.exec_one(Op("CALL", ["R12", "R13"]))
        assert mock_exec_call.call_count == 1
        assert mock_exec_call.call_args == (("R12", "R13"), {})


def test_call_changes_pc(vm):
    vm.pc = 100
    vm.registers[13] = 40
    vm.exec_call("R12", "R13")
    assert vm.pc == 40


def test_call_updates_second_register(vm):
    vm.pc = 100
    vm.registers[13] = 40
    vm.exec_call("R12", "R13")
    assert vm.registers[13] == 101


def test_call_updates_frame_pointer(vm):
    vm.registers[12] = 600
    vm.registers[13] = 40
    vm.exec_call("R12", "R13")
    assert vm.registers[14] == 600


def test_call_updates_first_register(vm):
    vm.registers[14] = 550
    vm.registers[12] = 600
    vm.registers[13] = 40
    vm.exec_call("R12", "R13")
    assert vm.registers[12] == 550


def test_exec_one_delegates_to_return(vm):
    with patch("hera.vm.VirtualMachine.exec_return") as mock_exec_return:
        vm.exec_one(Op("RETURN", ["R12", "R13"]))
        assert mock_exec_return.call_count == 1
        assert mock_exec_return.call_args == (("R12", "R13"), {})


def test_return_changes_pc(vm):
    vm.pc = 100
    vm.registers[13] = 40
    vm.exec_return("R12", "R13")
    assert vm.pc == 40


def test_return_updates_second_register(vm):
    vm.pc = 100
    vm.registers[13] = 40
    vm.exec_return("R12", "R13")
    assert vm.registers[13] == 101


def test_return_updates_frame_pointer(vm):
    vm.registers[12] = 600
    vm.registers[13] = 40
    vm.exec_return("R12", "R13")
    assert vm.registers[14] == 600


def test_return_updates_first_register(vm):
    vm.registers[14] = 550
    vm.registers[12] = 600
    vm.registers[13] = 40
    vm.exec_return("R12", "R13")
    assert vm.registers[12] == 550



"""Global configuration set by the main function."""


# Number of errors and warnings emitted.
ERROR_COUNT = 0
WARNING_COUNT = 0


def _make_ansi(*params):
    return "\033[" + ";".join(map(str, params)) + "m"


# ANSI color codes (https://stackoverflow.com/questions/4842424/)
# When the --no-color flag is specified, these constants are set to the empty string, so
# they can be used unconditionally in your code but will still obey the flag value.

ANSI_RED_BOLD = _make_ansi(31, 1)
ANSI_MAGENTA_BOLD = _make_ansi(35, 1)
ANSI_RESET = _make_ansi(0)



import pytest
from unittest.mock import patch

from hera.data import Op
from hera.utils import to_u16
from hera.vm import VirtualMachine


@pytest.fixture
def vm():
    return VirtualMachine()


def test_exec_one_delegates_to_load(vm):
    with patch("hera.vm.VirtualMachine.exec_load") as mock_exec_load:
        vm.exec_one(Op("LOAD", ["R1", 0, "R2"]))
        assert mock_exec_load.call_count == 1
        assert mock_exec_load.call_args == (("R1", 0, "R2"), {})


def test_load_from_small_address(vm):
    vm.memory[3] = 42
    vm.registers[2] = 3
    vm.exec_load("R1", 0, "R2")
    assert vm.registers[1] == 42


def test_load_from_uninitialized_address(vm):
    vm.registers[2] = 5
    vm.exec_load("R1", 0, "R2")
    assert vm.registers[1] == 0


def test_load_from_large_uninitialized_address(vm):
    vm.registers[2] = 14000
    vm.exec_load("R1", 0, "R2")
    assert vm.registers[1] == 0


def test_load_from_address_zero(vm):
    vm.memory[0] = 42
    vm.exec_load("R1", 0, "R0")
    assert vm.registers[1] == 42


def test_load_with_offset(vm):
    vm.memory[7] = 42
    vm.registers[2] = 4
    vm.exec_load("R1", 3, "R2")
    assert vm.registers[1] == 42


def test_load_sets_zero_flag(vm):
    vm.exec_load("R1", 0, "R0")
    assert vm.registers[1] == 0
    assert vm.flag_zero
    assert not vm.flag_sign


def test_load_sets_sign_flag(vm):
    vm.memory[4] = to_u16(-2)
    vm.registers[2] = 4
    vm.exec_load("R1", 0, "R2")
    assert vm.registers[1] == to_u16(-2)
    assert not vm.flag_zero
    assert vm.flag_sign


def test_load_ignores_other_flags(vm):
    vm.flag_carry = True
    vm.flag_overflow = True
    vm.flag_carry_block = True
    vm.exec_load("R1", 0, "R0")
    assert vm.flag_carry
    assert vm.flag_overflow
    assert vm.flag_carry_block


def test_load_does_not_affect_R0(vm):
    vm.memory[2] = 5
    vm.registers[2] = 2
    vm.exec_load("R0", 0, "R2")
    assert vm.registers[0] == 0


def test_load_increments_pc(vm):
    vm.exec_load("R1", 0, "R2")
    assert vm.pc == 1


def test_exec_one_delegates_to_store(vm):
    with patch("hera.vm.VirtualMachine.exec_store") as mock_exec_store:
        vm.exec_one(Op("STORE", ["R1", 0, "R2"]))
        assert mock_exec_store.call_count == 1
        assert mock_exec_store.call_args == (("R1", 0, "R2"), {})


def test_store_to_small_address(vm):
    vm.registers[1] = 42
    vm.registers[2] = 3
    vm.exec_store("R1", 0, "R2")
    assert vm.memory[3] == 42


def test_store_to_large_address(vm):
    vm.registers[1] = 42
    vm.registers[2] = 5000
    vm.exec_store("R1", 0, "R2")
    assert vm.memory[5000] == 42


def test_store_to_max_address(vm):
    vm.registers[1] = 42
    vm.registers[2] = (2 ** 16) - 1
    vm.exec_store("R1", 0, "R2")
    assert vm.memory[(2 ** 16) - 1] == 42


def test_store_to_address_zero(vm):
    vm.registers[1] = 42
    vm.exec_store("R1", 0, "R0")
    assert vm.memory[0] == 42


def test_store_with_offset(vm):
    vm.registers[1] = 42
    vm.registers[2] = 4
    vm.exec_store("R1", 3, "R2")
    assert vm.memory[7] == 42


def test_store_ignores_all_flags(vm):
    vm.flag_zero = True
    vm.flag_sign = True
    vm.flag_carry = True
    vm.flag_overflow = True
    vm.flag_carry_block = True
    vm.registers[1] = 42
    vm.exec_store("R1", 0, "R0")
    assert vm.flag_zero
    assert vm.flag_sign
    assert vm.flag_carry
    assert vm.flag_overflow
    assert vm.flag_carry_block


def test_store_increments_pc(vm):
    vm.exec_store("R1", 0, "R2")
    assert vm.pc == 1



from hera.main import main
from hera.symtab import HERA_DATA_START
from hera.vm import VirtualMachine


def test_strings_program():
    vm = VirtualMachine()
    main(["test/assets/manual/strings.hera"], vm)

    assert vm.registers[1] == 3
    assert vm.registers[2] == 0xC033
    assert vm.registers[3] == 0
    assert vm.registers[4] == 63
    assert vm.registers[5] == 63

    for r in vm.registers[6:10]:
        assert r == 0

    s = "Is this an example? With three questions? Really?"
    assert vm.memory[HERA_DATA_START] == len(s)
    for i in range(len(s)):
        assert vm.memory[HERA_DATA_START + i + 1] == ord(s[i])

    assert vm.flag_carry_block
    assert not vm.flag_carry
    assert not vm.flag_overflow
    assert vm.flag_zero
    assert not vm.flag_sign



"""The virtual HERA machine.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: December 2018
"""
import functools

from .symtab import HERA_DATA_START
from .utils import (
    DATA_STATEMENTS,
    emit_warning,
    from_u16,
    print_register_debug,
    register_to_index,
    to_u16,
    to_u32,
)


def ternary_op(f):
    """A decorator for ternary HERA ops. It handles fetching the values of the
    left and right registers, storing the result in the target register,
    setting the zero and sign flags, and incrementing the program counter.
    """

    @functools.wraps(f)
    def inner(self, target, left, right):
        left = self.get_register(left)
        right = self.get_register(right)
        result = f(self, left, right)
        self.store_register(target, result)
        self.set_zero_and_sign(result)
        self.pc += 1

    return inner


def binary_op(f):
    """A decorator for binary HERA ops. It handles fetching the value of the
    operand register, storing the result in the target register, setting the
    zero and sign flags, and incrementing the program counter.
    """

    @functools.wraps(f)
    def inner(self, target, original):
        original = self.get_register(original)
        result = f(self, original)
        self.store_register(target, result)
        self.set_zero_and_sign(result)
        self.pc += 1

    return inner


def branch(f):
    """A decorator for HERA register branching instructions. Implementing
    functions only need to return a boolean indicating whether to branch (True)
    or not (False).
    """

    @functools.wraps(f)
    def inner(self, dest):
        if f(self):
            self.pc = self.get_register(dest)
        else:
            self.pc += 1

    return inner


def relative_branch(f):
    """A decorator for HERA relative branching instructions. Implementing
    functions only need to return a boolean indicating whether to branch (True)
    or not (False).
    """

    @functools.wraps(f)
    def inner(self, offset):
        if f(self):
            self.pc += offset
        else:
            self.pc += 1

    return inner


class VirtualMachine:
    """An abstract representation of a HERA processor."""

    def __init__(self):
        self.reset()

    def reset(self):
        """Reset the machine to its initial state."""
        # Sixteen 16-bit registers. The virtual machine stores integers in their
        # unsigned representation, so the values of self.registers will always be
        # non-negative, although values above 2**15 implicitly represent negative
        # integers under a signed interpretation.
        self.registers = [0] * 16
        # 16-bit program counter
        self.pc = 0
        # Current memory cell for data instructions
        self.dc = HERA_DATA_START
        # Status/control flags
        self.flag_sign = False
        self.flag_zero = False
        self.flag_overflow = False
        self.flag_carry = False
        self.flag_carry_block = False
        # A memory array of 16-bit words. The HERA specification requires 2**16 words
        # to be addressable, but we start off with a considerably smaller array and
        # expand it as necessary, to keep the start-up time fast.
        self.memory = [0] * (2 ** 4)
        # Location object for the current operation
        self.location = None
        # Have warnings been issued for use of SWI and RTI instructions?
        self.warned_for_SWI = False
        self.warned_for_RTI = False

    def exec_one(self, op):
        """Execute a single operation."""
        self.location = getattr(op.name, "location", None)

        try:
            handler = getattr(self, "exec_" + op.name.lower())
        except AttributeError:
            raise RuntimeError('unknown instruction "{}"'.format(op.name)) from None
        else:
            handler(*op.args)

    def exec_many(self, program, *, lines=None):
        """Execute a program (i.e., a list of operations), resetting the machine's
        state beforehand.
        """
        self.reset()

        while program and program[0].name in DATA_STATEMENTS:
            data_statement = program.pop(0)
            self.exec_one(data_statement)

        self.pc = 0
        terminate_at = min(lines, len(program)) if lines is not None else len(program)
        while self.pc < terminate_at:
            opc = self.pc
            self.exec_one(program[self.pc])
            if opc == self.pc:
                break

    def get_register(self, name):
        """Get the contents of the register with the given name."""
        index = register_to_index(name)
        return self.registers[index]

    def store_register(self, target, value):
        """Store the value in the target register (a string)."""
        index = register_to_index(target)
        if index != 0:
            self.registers[index] = value

    def set_zero_and_sign(self, value):
        """Set the zero and sign flags based on the value."""
        self.flag_zero = value == 0
        self.flag_sign = value & 0x8000

    def assign_memory(self, address, value):
        """Assign a value to a location in memory."""
        # Extend the size of the memory array if necessary.
        if address >= len(self.memory):
            self.memory.extend([0] * (address - len(self.memory) + 1))
        self.memory[address] = value

    def access_memory(self, address):
        """Access a value in memory."""
        if address >= len(self.memory):
            return 0
        else:
            return self.memory[address]

    def exec_setlo(self, target, value):
        """Execute the SETLO instruction."""
        if value > 127:
            value -= 256

        self.store_register(target, to_u16(value))
        self.pc += 1

    def exec_sethi(self, target, value):
        """Execute the SETHI instruction."""
        self.store_register(target, (value << 8) + (self.get_register(target) & 0x00FF))
        self.pc += 1

    @ternary_op
    def exec_add(self, left, right):
        """Execute the ADD instruction."""
        carry = 1 if not self.flag_carry_block and self.flag_carry else 0

        result = (left + right + carry) & 0xFFFF

        self.flag_carry = result < (left + right + carry)
        self.flag_overflow = from_u16(result) != from_u16(left) + from_u16(right)

        return result

    @ternary_op
    def exec_sub(self, left, right):
        """Execute the SUB instruction."""
        borrow = 1 if not self.flag_carry_block and not self.flag_carry else 0

        # to_u16 is necessary because although left and right are necessarily
        # uints, left - right - borrow might not be.
        result = to_u16((left - right - borrow) & 0xFFFF)

        self.flag_carry = left >= right
        self.flag_overflow = (
            from_u16(result) != from_u16(left) - from_u16(right) - borrow
        )

        return result

    @ternary_op
    def exec_mul(self, left, right):
        """Execute the MUL instruction."""
        if self.flag_sign and not self.flag_carry_block:
            # Take the high 16 bits.
            left = to_u32(from_u16(left))
            right = to_u32(from_u16(right))
            result = ((left * right) & 0xFFFF0000) >> 16
        else:
            # Take the low 16 bits.
            result = (left * right) & 0xFFFF

        self.flag_carry = result < left * right
        self.flag_overflow = from_u16(result) != from_u16(left) * from_u16(right)

        return result

    @ternary_op
    def exec_and(self, left, right):
        """Execute the AND instruction."""
        return left & right

    @ternary_op
    def exec_or(self, left, right):
        """Execute the OR instruction."""
        return left | right

    @ternary_op
    def exec_xor(self, left, right):
        """Execute the XOR instruction."""
        return left ^ right

    def exec_inc(self, target, value):
        """Execute the INC (increment) instruction."""
        original = self.get_register(target)
        result = (value + original) & 0xFFFF
        self.store_register(target, result)

        self.set_zero_and_sign(result)
        self.flag_overflow = from_u16(result) != from_u16(original) + value
        self.flag_carry = value + original >= 2 ** 16
        self.pc += 1

    def exec_dec(self, target, value):
        """Execute the DEC (decrement) instruction."""
        original = self.get_register(target)
        result = to_u16((original - value) & 0xFFFF)
        self.store_register(target, result)

        self.set_zero_and_sign(result)
        self.flag_overflow = from_u16(result) != from_u16(original) - value
        self.flag_carry = original < value
        self.pc += 1

    @binary_op
    def exec_lsl(self, original):
        """Execute the LSL (logical shift left) instruction."""
        carry = 1 if self.flag_carry and not self.flag_carry_block else 0
        result = ((original << 1) + carry) & 0xFFFF

        self.flag_carry = original & 0x8000

        return result

    @binary_op
    def exec_lsr(self, original):
        """Execute the LSR (logical shift right) instruction."""
        carry = 2 ** 15 if self.flag_carry and not self.flag_carry_block else 0
        result = (original >> 1) + carry

        self.flag_carry = original % 2 == 1

        return result

    @binary_op
    def exec_lsl8(self, original):
        """Execute the LSL8 (logical shift left by 8) instruction."""
        return (original << 8) & 0xFFFF

    @binary_op
    def exec_lsr8(self, original):
        """Execute the LSR8 (logical shift right by 8) instruction."""
        return original >> 8

    @binary_op
    def exec_asl(self, original):
        """Execute the ASL (arithmetic shift left) instruction."""
        carry = 1 if self.flag_carry and not self.flag_carry_block else 0
        result = ((original << 1) + carry) & 0xFFFF

        self.flag_carry = original & 0x8000
        self.flag_overflow = original & 0x8000 and not result & 0x8000

        return result

    @binary_op
    def exec_asr(self, original):
        """Execute the ASR (arithmetic shift right) instruction."""
        # This is a little messy because right shift in Python rounds towards
        # negative infinity (7 >> 1 == -4) but in HERA it rounds towards zero
        # (7 >> 1 == -3).
        if original & 0x8000:
            if original & 0x0001:
                result = ((original >> 1) | 0x8000) + 1
            else:
                result = original >> 1 | 0x8000
        else:
            result = original >> 1

        self.flag_carry = original & 0x0001

        return result

    def exec_savef(self, target):
        """Execute the SAVEF (save flags) instruction."""
        value = (
            int(self.flag_sign)
            + 2 * int(self.flag_zero)
            + 4 * int(self.flag_overflow)
            + 8 * int(self.flag_carry)
            + 16 * int(self.flag_carry_block)
        )
        self.store_register(target, value)
        self.pc += 1

    def exec_rstrf(self, target):
        """Execute the RSTRF (restore flags) instruction."""
        value = self.get_register(target)
        self.flag_sign = bool(value & 1)
        self.flag_zero = bool(value & 0b10)
        self.flag_overflow = bool(value & 0b100)
        self.flag_carry = bool(value & 0b1000)
        self.flag_carry_block = bool(value & 0b10000)
        self.pc += 1

    def exec_fon(self, value):
        """Execute the FON instruction."""
        self.flag_sign = self.flag_sign or bool(value & 1)
        self.flag_zero = self.flag_zero or bool(value & 0b10)
        self.flag_overflow = self.flag_overflow or bool(value & 0b100)
        self.flag_carry = self.flag_carry or bool(value & 0b1000)
        self.flag_carry_block = self.flag_carry_block or bool(value & 0b10000)
        self.pc += 1

    def exec_foff(self, value):
        """Execute the FOFF instruction."""
        self.flag_sign = self.flag_sign and not bool(value & 1)
        self.flag_zero = self.flag_zero and not bool(value & 0b10)
        self.flag_overflow = self.flag_overflow and not bool(value & 0b100)
        self.flag_carry = self.flag_carry and not bool(value & 0b1000)
        self.flag_carry_block = self.flag_carry_block and not bool(value & 0b10000)
        self.pc += 1

    def exec_fset5(self, value):
        """Execute the FSET5 instruction."""
        self.flag_sign = bool(value & 1)
        self.flag_zero = bool(value & 0b10)
        self.flag_overflow = bool(value & 0b100)
        self.flag_carry = bool(value & 0b1000)
        self.flag_carry_block = bool(value & 0b10000)
        self.pc += 1

    def exec_fset4(self, value):
        """Execute the FSET4 instruction."""
        self.flag_sign = bool(value & 1)
        self.flag_zero = bool(value & 0b10)
        self.flag_overflow = bool(value & 0b100)
        self.flag_carry = bool(value & 0b1000)
        self.pc += 1

    def exec_load(self, target, offset, address):
        """Execute the LOAD instruction."""
        result = self.access_memory(self.get_register(address) + offset)
        self.set_zero_and_sign(result)
        self.store_register(target, result)
        self.pc += 1

    def exec_store(self, source, offset, address):
        """Execute the STORE instruction."""
        self.assign_memory(
            self.get_register(address) + offset, self.get_register(source)
        )
        self.pc += 1

    def exec_br(self, dest):
        """Execute the BR (unconditional branch) instruction."""
        self.pc = self.get_register(dest)

    def exec_brr(self, offset):
        """Execute the BRR (unconditional branch) instruction."""
        self.pc += offset

    @branch
    def exec_bl(self):
        """Execute the BL (branch if less than) instruction."""
        return self.flag_sign ^ self.flag_overflow

    @relative_branch
    def exec_blr(self):
        """Execute the BLR (branch if less than) instruction."""
        return self.flag_sign ^ self.flag_overflow

    @branch
    def exec_bge(self):
        """Execute the BGE (branch if greater than or equal) instruction."""
        return not (self.flag_sign ^ self.flag_overflow)

    @relative_branch
    def exec_bger(self):
        """Execute the BGER (branch if greater than or equal) instruction."""
        return not (self.flag_sign ^ self.flag_overflow)

    @branch
    def exec_ble(self):
        """Execute the BLE (branch if less than or equal) instruction."""
        return (self.flag_sign ^ self.flag_overflow) or self.flag_zero

    @relative_branch
    def exec_bler(self):
        """Execute the BLER (branch if less than or equal) instruction."""
        return (self.flag_sign ^ self.flag_overflow) or self.flag_zero

    @branch
    def exec_bg(self):
        """Execute the BG (branch if greater than) instruction."""
        return not (self.flag_sign ^ self.flag_overflow) and not self.flag_zero

    @relative_branch
    def exec_bgr(self):
        """Execute the BGR (branch if greater than) instruction."""
        return not (self.flag_sign ^ self.flag_overflow) and not self.flag_zero

    @branch
    def exec_bule(self):
        """Execute the BULE (branch if unsigned less than or equal) instruction."""
        return not self.flag_carry or self.flag_zero

    @relative_branch
    def exec_buler(self):
        """Execute the BULER (branch if unsigned less than or equal) instruction."""
        return not self.flag_carry or self.flag_zero

    @branch
    def exec_bug(self):
        """Execute the BUG (branch if unsigned greater than) instruction."""
        return self.flag_carry and not self.flag_zero

    @relative_branch
    def exec_bugr(self):
        """Execute the BUGR (branch if unsigned greater than) instruction."""
        return self.flag_carry and not self.flag_zero

    @branch
    def exec_bz(self):
        """Execute the BZ (branch on zero flag) instruction."""
        return self.flag_zero

    @relative_branch
    def exec_bzr(self):
        """Execute the BZR (branch on zero flag) instruction."""
        return self.flag_zero

    @branch
    def exec_bnz(self):
        """Execute the BNZ (branch on not zero flag) instruction."""
        return not self.flag_zero

    @relative_branch
    def exec_bnzr(self):
        """Execute the BNZR (branch on not zero flag) instruction."""
        return not self.flag_zero

    @branch
    def exec_bc(self):
        """Execute the BC (branch on carry flag) instruction."""
        return self.flag_carry

    @relative_branch
    def exec_bcr(self):
        """Execute the BCR (branch on carry flag) instruction."""
        return self.flag_carry

    @branch
    def exec_bnc(self):
        """Execute the BNC (branch on not carry flag) instruction."""
        return not self.flag_carry

    @relative_branch
    def exec_bncr(self):
        """Execute the BNCR (branch on not carry flag) instruction."""
        return not self.flag_carry

    @branch
    def exec_bs(self):
        """Execute the BS (branch on sign flag) instruction."""
        return self.flag_sign

    @relative_branch
    def exec_bsr(self):
        """Execute the BSR (branch on sign flag) instruction."""
        return self.flag_sign

    @branch
    def exec_bns(self):
        """Execute the BNS (branch on not sign flag) instruction."""
        return not self.flag_sign

    @relative_branch
    def exec_bnsr(self):
        """Execute the BNSR (branch on not sign flag) instruction."""
        return not self.flag_sign

    @branch
    def exec_bv(self):
        """Execute the BV (branch on overflow flag) instruction."""
        return self.flag_overflow

    @relative_branch
    def exec_bvr(self):
        """Execute the BVR (branch on overflow flag) instruction."""
        return self.flag_overflow

    @branch
    def exec_bnv(self):
        """Execute the BNV (branch on not overflow flag) instruction."""
        return not self.flag_overflow

    @relative_branch
    def exec_bnvr(self):
        """Execute the BNVR (branch on not overflow flag) instruction."""
        return not self.flag_overflow

    def exec_call(self, ra, rb):
        """Execute the CALL instruction."""
        old_pc = self.pc
        self.pc = self.get_register(rb)
        self.store_register(rb, old_pc + 1)
        old_fp = self.get_register("FP")
        self.store_register("FP", self.get_register(ra))
        self.store_register(ra, old_fp)

    # CALL and RETURN do the exact same thing.
    exec_return = exec_call

    def exec_swi(self, i):
        """Execute the SWI (software interrupt) instruction."""
        if not self.warned_for_SWI:
            emit_warning("SWI is a no-op in this simulator", loc=self.location)
            self.warned_for_SWI = True
        self.pc += 1

    def exec_rti(self):
        """Execute the RTI (return from interrupt) instruction."""
        if not self.warned_for_RTI:
            emit_warning("RTI is a no-op in this simulator", loc=self.location)
            self.warned_for_RTI = True
        self.pc += 1

    def exec_integer(self, i):
        """Execute the INTEGER data instruction."""
        self.assign_memory(self.dc, to_u16(i))
        self.dc += 1
        self.pc += 1

    def exec_dskip(self, n):
        """Execute the DSKIP data instruction."""
        self.dc += n
        self.pc += 1

    def exec_lp_string(self, s):
        """Execute the LP_STRING data instruction."""
        self.assign_memory(self.dc, len(s))
        self.dc += 1
        for c in s:
            self.assign_memory(self.dc, ord(c))
            self.dc += 1
        self.pc += 1

    def exec_print_reg(self, target):
        """Execute the print_reg debugging instruction."""
        v = self.get_register(target)
        print_register_debug(target, v, to_stderr=False)
        self.pc += 1

    def exec_print(self, target):
        """Execute the print debugging instruction."""
        print(target, end="")
        self.pc += 1

    def exec_println(self, target):
        """Execute the println debugging instruction."""
        print(target)
        self.pc += 1



from hera.data import Op, Token
from hera.preprocessor import convert, convert_set, preprocess, substitute_label


def R(s):
    return Token("REGISTER", s)


def SYM(s):
    return Token("SYMBOL", s)


def test_substitute_label_with_SETLO():
    labels = {"N": 10}
    assert substitute_label(Op(SYM("SETLO"), [R("R1"), SYM("N")]), labels) == Op(
        "SETLO", ["R1", 10]
    )


def test_substitute_label_with_SETHI():
    labels = {"N": 10}
    assert substitute_label(Op(SYM("SETHI"), [R("R1"), SYM("N")]), labels) == Op(
        "SETHI", ["R1", 10]
    )


def test_substitute_label_with_other_op():
    labels = {"N": 10}
    assert substitute_label(Op(SYM("INC"), [R("R1"), SYM("N")]), labels) == Op(
        "INC", ["R1", 10]
    )


def test_convert_set_with_small_positive():
    assert convert_set("R5", 18) == [Op("SETLO", ["R5", 18]), Op("SETHI", ["R5", 0])]


def test_convert_set_with_large_positive():
    assert convert_set("R5", 34000) == [
        Op("SETLO", ["R5", 208]),
        Op("SETHI", ["R5", 132]),
    ]


def test_convert_set_with_negative():
    assert convert_set("R5", -5) == [Op("SETLO", ["R5", 251]), Op("SETHI", ["R5", 255])]


def test_convert_set_with_symbol():
    assert convert_set("R5", "whatever") == [
        Op("SETLO", ["R5", "whatever"]),
        Op("SETHI", ["R5", "whatever"]),
    ]


def test_convert_move():
    assert convert(Op("MOVE", ["R5", "R3"])) == [Op("OR", ["R5", "R3", "R0"])]


def test_convert_con():
    assert convert(Op("CON", [])) == [Op("FON", [8])]


def test_convert_coff():
    assert convert(Op("COFF", [])) == [Op("FOFF", [8])]


def test_convert_cbon():
    assert convert(Op("CBON", [])) == [Op("FON", [16])]


def test_convert_ccboff():
    assert convert(Op("CCBOFF", [])) == [Op("FOFF", [24])]


def test_convert_cmp():
    assert convert(Op("CMP", ["R1", "R2"])) == [
        Op("FON", [8]),
        Op("SUB", ["R0", "R1", "R2"]),
    ]


def test_convert_setrf_with_small_positive():
    assert convert(Op("SETRF", ["R5", 18])) == [
        Op("SETLO", ["R5", 18]),
        Op("SETHI", ["R5", 0]),
        Op("FOFF", [8]),
        Op("ADD", ["R0", "R5", "R0"]),
    ]


def test_convert_setrf_with_large_positive():
    assert convert(Op("SETRF", ["R5", 34000])) == [
        Op("SETLO", ["R5", 208]),
        Op("SETHI", ["R5", 132]),
        Op("FOFF", [8]),
        Op("ADD", ["R0", "R5", "R0"]),
    ]


def test_convert_setrf_with_negative():
    assert convert(Op("SETRF", ["R5", -5])) == [
        Op("SETLO", ["R5", 251]),
        Op("SETHI", ["R5", 255]),
        Op("FOFF", [8]),
        Op("ADD", ["R0", "R5", "R0"]),
    ]


def test_convert_flags():
    assert convert(Op("FLAGS", ["R8"])) == [
        Op("FOFF", [8]),
        Op("ADD", ["R0", "R8", "R0"]),
    ]


def test_convert_halt():
    assert convert(Op("HALT", [])) == [Op("BRR", [0])]


def test_convert_nop():
    assert convert(Op("NOP", [])) == [Op("BRR", [1])]


def test_convert_call_with_register():
    assert convert(Op("CALL", ["R12", R("R13")])) == [Op("CALL", ["R12", "R13"])]


def test_convert_call_with_label():
    assert convert(Op("CALL", ["R12", Token("SYMBOL", "div")])) == [
        Op("SETLO", ["R13", "div"]),
        Op("SETHI", ["R13", "div"]),
        Op("CALL", ["R12", "R13"]),
    ]


def test_convert_neg():
    assert convert(Op("NEG", ["R1", "R2"])) == [
        Op("FON", [8]),
        Op("SUB", ["R1", "R0", "R2"]),
    ]


def test_convert_not():
    assert convert(Op("NOT", ["R1", "R2"])) == [
        Op("SETLO", ["R11", 0xFF]),
        Op("SETHI", ["R11", 0xFF]),
        Op("XOR", ["R1", "R11", "R2"]),
    ]


def test_preprocess_constant():
    program = [Op(Token("SYMBOL", "SET"), [R("R1"), Token("SYMBOL", "n")])]
    assert preprocess(program, {"n": 100}) == [
        Op("SETLO", ["R1", 100]),
        Op("SETHI", ["R1", 0]),
    ]



from hera.data import Op, Token
from hera.symtab import get_symtab, HERA_DATA_START


def SYM(s):
    return Token("SYMBOL", s)


def test_get_symtab_with_example():
    labels = get_symtab(
        [
            Op("DLABEL", ["data"]),
            Op("INTEGER", [42]),
            Op("INTEGER", [43]),
            Op("DLABEL", ["data2"]),
            Op("INTEGER", [100]),
            Op("LABEL", ["top"]),
            Op("ADD", ["R0", "R0", "R0"]),
            Op("LABEL", ["bottom"]),
        ]
    )
    assert len(labels) == 4
    assert labels["data"] == HERA_DATA_START
    assert labels["data2"] == HERA_DATA_START + 2
    assert labels["top"] == 0
    assert labels["bottom"] == 1


def test_get_symtab_with_dskip():
    labels = get_symtab(
        [
            Op("DLABEL", ["data"]),
            Op("INTEGER", [42]),
            Op("DSKIP", [10]),
            Op("DLABEL", ["data2"]),
            Op("INTEGER", [84]),
        ]
    )
    assert len(labels) == 2
    assert labels["data"] == HERA_DATA_START
    assert labels["data2"] == HERA_DATA_START + 11


def test_get_symtab_with_dskip_and_constant():
    labels = get_symtab(
        [
            Op("CONSTANT", ["N", 50]),
            Op("DLABEL", ["x"]),
            Op("DSKIP", ["N"]),  # Using constant value in DSKIP.
            Op("DLABEL", ["y"]),
        ]
    )
    assert len(labels) == 3
    assert labels["N"] == 50
    assert labels["x"] == HERA_DATA_START
    assert labels["y"] == HERA_DATA_START + 50


def test_get_symtab_with_lp_string():
    labels = get_symtab(
        [
            Op("DLABEL", ["S"]),
            Op("LP_STRING", ["hello"]),
            Op("DLABEL", ["X"]),
            Op("INTEGER", [42]),
        ]
    )
    assert len(labels) == 2
    assert labels["S"] == HERA_DATA_START
    assert labels["X"] == HERA_DATA_START + 6


def test_get_symtab_with_empty_lp_string():
    labels = get_symtab(
        [
            Op("DLABEL", ["S"]),
            Op("LP_STRING", [""]),
            Op("DLABEL", ["X"]),
            Op("INTEGER", [42]),
        ]
    )
    assert len(labels) == 2
    assert labels["S"] == HERA_DATA_START
    assert labels["X"] == HERA_DATA_START + 1


def test_get_symtab_with_invalid_instructions():
    labels = get_symtab([Op("CONSTANT", ["N"]), Op("CONSTANT", ["X", 42])])
    assert len(labels) == 1
    assert labels["X"] == 42


def test_get_symtab_with_too_large_dskip(capsys):
    get_symtab([Op(SYM("DSKIP"), [1000000000])])

    assert "past the end of available memory" in capsys.readouterr().err


def test_get_symtab_with_redefinitions_of_symbols(capsys):
    get_symtab(
        [
            Op(SYM("CONSTANT"), ["A", 100]),
            Op(SYM("CONSTANT"), ["B", 200]),
            Op(SYM("CONSTANT"), ["C", 200]),
            Op(SYM("CONSTANT"), ["A", -1]),
            Op(SYM("LABEL"), ["B"]),
            Op(SYM("DLABEL"), ["C"]),
        ]
    )

    captured = capsys.readouterr().err

    assert "symbol `A` has already been defined" in captured
    assert "symbol `B` has already been defined" in captured
    assert "symbol `C` has already been defined" in captured



import pytest
from unittest.mock import patch

from hera.data import Op
from hera.utils import to_u16
from hera.vm import VirtualMachine


@pytest.fixture
def vm():
    return VirtualMachine()


def test_exec_one_delegates_to_add(vm):
    with patch("hera.vm.VirtualMachine.exec_add") as mock_exec_add:
        vm.exec_one(Op("ADD", ["R1", "R2", "R3"]))
        assert mock_exec_add.call_count == 1
        assert mock_exec_add.call_args == (("R1", "R2", "R3"), {})


def test_add_small_numbers(vm):
    vm.registers[2] = 20
    vm.registers[3] = 22
    vm.exec_add("R1", "R2", "R3")
    assert vm.registers[1] == 42


def test_add_increments_pc(vm):
    vm.exec_add("R1", "R2", "R3")
    assert vm.pc == 1


def test_add_sets_flags(vm):
    vm.registers[2] = 20
    vm.registers[3] = 22
    vm.exec_add("R1", "R2", "R3")
    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry


def test_add_with_negative(vm):
    vm.registers[2] = to_u16(-14)
    vm.registers[3] = 8
    vm.exec_add("R1", "R2", "R3")
    assert vm.registers[1] == to_u16(-6)
    assert vm.flag_sign
    assert not vm.flag_zero


def test_add_with_zero(vm):
    vm.registers[7] = to_u16(-4)
    vm.registers[3] = 4
    vm.exec_add("R5", "R7", "R3")
    assert vm.registers[5] == 0
    assert not vm.flag_sign
    assert vm.flag_zero


def test_add_with_overflow(vm):
    vm.registers[9] = 32767
    vm.registers[2] = 1
    vm.exec_add("R7", "R9", "R2")
    assert vm.registers[7] == to_u16(-32768)
    assert vm.flag_sign
    assert vm.flag_overflow
    assert not vm.flag_carry


def test_add_with_big_overflow(vm):
    vm.registers[9] = 32767
    vm.registers[2] = 32767
    vm.exec_add("R7", "R9", "R2")
    assert vm.registers[7] == to_u16(-2)
    assert vm.flag_sign
    assert vm.flag_overflow
    assert not vm.flag_carry


def test_add_with_negative_overflow(vm):
    vm.registers[9] = to_u16(-32768)
    vm.registers[2] = to_u16(-32768)
    vm.exec_add("R7", "R9", "R2")
    assert vm.registers[7] == 0
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry


def test_add_with_carry(vm):
    vm.registers[3] = 5
    vm.registers[5] = 3
    vm.flag_carry = True
    vm.exec_add("R7", "R3", "R5")
    assert vm.registers[7] == 9
    assert not vm.flag_carry


def test_add_with_carry_and_block(vm):
    vm.registers[3] = 5
    vm.registers[5] = 3
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_add("R7", "R3", "R5")
    assert vm.registers[7] == 8
    assert not vm.flag_carry
    assert vm.flag_carry_block


def test_add_with_overflow_from_carry(vm):
    vm.registers[2] = 32760
    vm.registers[3] = 7
    vm.flag_carry = True
    vm.exec_add("R1", "R2", "R3")
    assert vm.registers[1] == to_u16(-32768)
    assert vm.flag_sign
    assert vm.flag_overflow
    assert not vm.flag_carry


def test_add_does_not_change_R0(vm):
    vm.registers[1] = 1
    vm.registers[2] = 1
    vm.exec_add("R0", "R1", "R2")
    assert vm.registers[0] == 0


def test_exec_one_delegates_to_sub(vm):
    with patch("hera.vm.VirtualMachine.exec_sub") as mock_exec_sub:
        vm.exec_one(Op("SUB", ["R1", "R2", "R3"]))
        assert mock_exec_sub.call_count == 1
        assert mock_exec_sub.call_args == (("R1", "R2", "R3"), {})


def test_sub_small_numbers(vm):
    vm.registers[2] = 64
    vm.registers[3] = 22
    vm.flag_carry_block = True
    vm.exec_sub("R1", "R2", "R3")
    assert vm.registers[1] == 42


def test_sub_sets_flags(vm):
    vm.registers[2] = 64
    vm.registers[3] = 22
    vm.flag_carry_block = True
    vm.exec_sub("R1", "R2", "R3")
    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry


def test_sub_increments_pc(vm):
    vm.exec_sub("R1", "R2", "R3")
    assert vm.pc == 1


def test_sub_with_negative(vm):
    vm.registers[2] = to_u16(-64)
    vm.registers[3] = 22
    vm.flag_carry_block = True
    vm.exec_sub("R1", "R2", "R3")
    assert vm.registers[1] == to_u16(-86)
    assert vm.flag_sign
    assert not vm.flag_zero


def test_sub_with_zero(vm):
    vm.registers[2] = to_u16(-37)
    vm.registers[3] = to_u16(-37)
    vm.flag_carry_block = True
    vm.exec_sub("R1", "R2", "R3")
    assert vm.registers[1] == 0
    assert not vm.flag_sign
    assert vm.flag_zero


def test_sub_with_two_negatives(vm):
    vm.registers[2] = to_u16(-20)
    vm.registers[3] = to_u16(-40)
    vm.flag_carry_block = True
    vm.exec_sub("R1", "R2", "R3")
    assert vm.registers[1] == 20
    assert not vm.flag_sign
    assert vm.flag_carry
    assert not vm.flag_overflow


def test_sub_with_min_negative_overflow(vm):
    vm.registers[1] = to_u16(-32768)
    vm.registers[2] = 1
    vm.flag_carry_block = True
    vm.exec_sub("R3", "R1", "R2")
    assert vm.registers[3] == 32767
    assert not vm.flag_sign
    assert vm.flag_carry
    assert vm.flag_overflow


def test_sub_with_big_negative_overflow(vm):
    vm.registers[1] = to_u16(-32000)
    vm.registers[2] = 32000
    vm.flag_carry_block = True
    vm.exec_sub("R3", "R1", "R2")
    assert vm.registers[3] == 1536
    assert not vm.flag_sign
    assert vm.flag_carry
    assert vm.flag_overflow


def test_sub_with_max_negative_overflow(vm):
    vm.registers[1] = to_u16(-32768)
    vm.registers[2] = 32767
    vm.flag_carry_block = True
    vm.exec_sub("R3", "R1", "R2")
    assert vm.registers[3] == 1
    assert not vm.flag_sign
    assert vm.flag_carry
    assert vm.flag_overflow


def test_sub_with_min_positive_overflow(vm):
    vm.registers[4] = 32767
    vm.registers[5] = to_u16(-1)
    vm.flag_carry_block = True
    vm.exec_sub("R6", "R4", "R5")
    assert vm.registers[6] == to_u16(-32768)
    assert vm.flag_sign
    assert not vm.flag_carry
    assert vm.flag_overflow


def test_sub_with_big_positive_overflow(vm):
    vm.registers[4] = 27500
    vm.registers[5] = to_u16(-7040)
    vm.flag_carry_block = True
    vm.exec_sub("R6", "R4", "R5")
    assert vm.registers[6] == to_u16(-30996)
    assert vm.flag_sign
    assert not vm.flag_carry
    assert vm.flag_overflow


def test_sub_with_max_positive_overflow(vm):
    vm.registers[4] = 32767
    vm.registers[5] = to_u16(-32768)
    vm.flag_carry_block = True
    vm.exec_sub("R6", "R4", "R5")
    assert vm.registers[6] == to_u16(-1)
    assert vm.flag_sign
    assert not vm.flag_carry
    assert vm.flag_overflow


def test_sub_with_implicit_borrow(vm):
    vm.registers[2] = 17
    vm.registers[3] = 5
    vm.exec_sub("R1", "R2", "R3")
    assert vm.registers[1] == 11


def test_sub_with_no_carry_block_and_no_borrow(vm):
    vm.registers[2] = to_u16(-64)
    vm.registers[3] = 22
    vm.flag_carry = True
    vm.exec_sub("R1", "R2", "R3")
    assert vm.registers[1] == to_u16(-86)
    assert vm.flag_carry


def test_sub_overflow_from_borrow(vm):
    vm.registers[1] = to_u16(-32767)
    vm.registers[2] = 1
    vm.exec_sub("R3", "R1", "R2")
    assert vm.registers[3] == 32767
    assert not vm.flag_sign
    assert vm.flag_carry
    assert vm.flag_overflow


def test_sub_overflow_takes_borrow_into_account(vm):
    vm.registers[1] = 10
    vm.registers[2] = 11
    vm.exec_sub("R0", "R1", "R2")
    assert not vm.flag_overflow


def test_sub_sets_carry_for_equal_operands(vm):
    vm.flag_carry = True
    vm.registers[1] = 12
    vm.registers[2] = 12
    vm.exec_sub("R3", "R1", "R2")
    assert vm.registers[3] == 0
    assert vm.flag_carry


def test_sub_does_not_affect_R0(vm):
    vm.registers[1] = 4
    vm.registers[2] = 3
    vm.exec_sub("R0", "R1", "R2")
    assert vm.registers[0] == 0


def test_exec_one_delegates_to_mul(vm):
    with patch("hera.vm.VirtualMachine.exec_mul") as mock_exec_mul:
        vm.exec_one(Op("MUL", ["R1", "R2", "R3"]))
        assert mock_exec_mul.call_count == 1
        assert mock_exec_mul.call_args == (("R1", "R2", "R3"), {})


def test_mul_with_small_positives(vm):
    vm.registers[2] = 4
    vm.registers[3] = 2
    vm.exec_mul("R1", "R2", "R3")
    assert vm.registers[1] == 8


def test_mul_with_large_positives(vm):
    vm.registers[2] = 4500
    vm.registers[3] = 3
    vm.exec_mul("R1", "R2", "R3")
    assert vm.registers[1] == 13500


def test_mul_with_small_negatives(vm):
    vm.registers[2] = to_u16(-5)
    vm.registers[3] = 3
    vm.exec_mul("R1", "R2", "R3")
    assert vm.registers[1] == to_u16(-15)


def test_mul_with_large_negatives(vm):
    vm.registers[2] = 7
    vm.registers[3] = to_u16(-3100)
    vm.exec_mul("R1", "R2", "R3")
    assert vm.registers[1] == to_u16(-21700)


def test_mul_sets_zero_flag(vm):
    vm.registers[2] = 7
    vm.registers[3] = 0
    vm.exec_mul("R1", "R2", "R3")
    assert vm.registers[1] == 0
    assert vm.flag_zero
    assert not vm.flag_sign


def test_mul_sets_sign_flag(vm):
    vm.registers[2] = to_u16(-1)
    vm.registers[3] = 17
    vm.exec_mul("R1", "R2", "R3")
    assert vm.registers[1] == to_u16(-17)
    assert not vm.flag_zero
    assert vm.flag_sign


def test_mul_with_signed_positive_overflow(vm):
    vm.registers[2] = 17000
    vm.registers[3] = 3
    vm.exec_mul("R1", "R2", "R3")
    assert vm.registers[1] == to_u16(-14536)
    assert not vm.flag_carry
    assert vm.flag_overflow


def test_mul_with_unsigned_positive_overflow(vm):
    vm.registers[2] = 128
    vm.registers[3] = 749
    vm.exec_mul("R1", "R2", "R3")
    assert vm.registers[1] == 30336
    assert vm.flag_carry
    assert vm.flag_overflow


def test_mul_with_signed_negative_overflow(vm):
    vm.registers[2] = to_u16(-400)
    vm.registers[3] = to_u16(-200)
    vm.exec_mul("R1", "R2", "R3")
    assert vm.registers[1] == 14464
    assert vm.flag_carry
    assert vm.flag_overflow


def test_mul_ignores_carry_when_blocked(vm):
    vm.flag_carry_block = True
    vm.flag_carry = True
    vm.registers[2] = 4
    vm.registers[3] = 12
    vm.exec_mul("R1", "R2", "R3")
    assert vm.registers[1] == 48
    assert not vm.flag_carry


def test_mul_ignores_carry_when_not_blocked(vm):
    vm.flag_carry = True
    vm.registers[2] = 4
    vm.registers[3] = 12
    vm.exec_mul("R1", "R2", "R3")
    assert vm.registers[1] == 48
    assert not vm.flag_carry


def test_mul_produces_high_bits_when_sign_flag_is_on(vm):
    vm.flag_sign = True
    vm.registers[2] = 20000
    vm.registers[3] = 200
    vm.exec_mul("R1", "R2", "R3")
    assert vm.registers[1] == 0b111101
    assert not vm.flag_sign
    # TODO: Find out how the carry and overflow flags should be set.


def test_mul_with_sign_flag_and_negative_result(vm):
    vm.flag_sign = True
    vm.registers[2] = 20000
    vm.registers[3] = to_u16(-200)
    vm.exec_mul("R1", "R2", "R3")
    assert vm.registers[1] == 0b1111111111000010
    assert vm.flag_sign


def test_mul_with_sign_flag_zero_result(vm):
    vm.flag_sign = True
    vm.registers[2] = 47
    vm.registers[3] = 0
    vm.exec_mul("R1", "R2", "R3")
    assert vm.registers[1] == 0
    assert not vm.flag_sign
    assert vm.flag_zero


def test_mul_ignores_sign_flag_when_carry_is_blocked(vm):
    vm.flag_sign = True
    vm.flag_carry_block = True
    vm.registers[2] = 20000
    vm.registers[3] = 200
    vm.exec_mul("R1", "R2", "R3")
    assert vm.registers[1] == 2304
    assert not vm.flag_sign
    assert vm.flag_carry


def test_mul_increments_pc(vm):
    vm.exec_mul("R1", "R2", "R3")
    assert vm.pc == 1


def test_mul_does_not_affect_R0(vm):
    vm.registers[1] = 4
    vm.registers[2] = 3
    vm.exec_mul("R0", "R1", "R2")
    assert vm.registers[0] == 0


def test_exec_one_delegates_to_and(vm):
    with patch("hera.vm.VirtualMachine.exec_and") as mock_exec_and:
        vm.exec_one(Op("AND", ["R1", "R2", "R3"]))
        assert mock_exec_and.call_count == 1
        assert mock_exec_and.call_args == (("R1", "R2", "R3"), {})


def test_and_same_numbers(vm):
    vm.registers[2] = 27
    vm.registers[3] = 27
    vm.exec_and("R1", "R2", "R3")
    assert vm.registers[1] == 27


def test_and_different_numbers(vm):
    vm.registers[2] = 3  # 011
    vm.registers[3] = 6  # 110
    vm.exec_and("R1", "R2", "R3")
    assert vm.registers[1] == 2


def test_and_increments_pc(vm):
    vm.exec_and("R0", "R1", "R2")
    assert vm.pc == 1


def test_and_big_numbers(vm):
    vm.registers[2] = 62434
    vm.registers[3] = 17589
    vm.exec_and("R1", "R2", "R3")
    assert vm.registers[1] == 16544


def test_and_sets_zero_flag(vm):
    vm.registers[2] = 82
    vm.registers[3] = 0
    vm.exec_and("R1", "R2", "R3")
    assert vm.registers[1] == 0
    assert vm.flag_zero
    assert not vm.flag_sign


def test_and_sets_sign_flag(vm):
    vm.registers[2] = to_u16(-1)
    vm.registers[3] = to_u16(-37)
    vm.exec_and("R1", "R2", "R3")
    assert vm.registers[1] == to_u16(-37)
    assert not vm.flag_zero
    assert vm.flag_sign


def test_and_does_not_set_other_flags(vm):
    vm.registers[2] = to_u16(-1)
    vm.registers[3] = to_u16(-1)
    vm.exec_and("R1", "R2", "R3")
    assert vm.registers[1] == to_u16(-1)
    assert not vm.flag_carry
    assert not vm.flag_overflow


def test_and_does_not_clear_other_flags(vm):
    vm.registers[2] = to_u16(-1)
    vm.registers[3] = to_u16(-1)
    vm.flag_carry = True
    vm.flag_overflow = True
    vm.exec_and("R1", "R2", "R3")
    assert vm.flag_carry
    assert vm.flag_overflow


def test_and_does_not_affect_R0(vm):
    vm.registers[1] = 1
    vm.registers[2] = 1
    vm.exec_and("R0", "R1", "R2")
    assert vm.registers[0] == 0


def test_exec_one_delegates_to_or(vm):
    with patch("hera.vm.VirtualMachine.exec_or") as mock_exec_or:
        vm.exec_one(Op("OR", ["R1", "R2", "R3"]))
        assert mock_exec_or.call_count == 1
        assert mock_exec_or.call_args == (("R1", "R2", "R3"), {})


def test_or_same_numbers(vm):
    vm.registers[2] = 27
    vm.registers[3] = 27
    vm.exec_or("R1", "R2", "R3")
    assert vm.registers[1] == 27


def test_or_different_numbers(vm):
    vm.registers[2] = 3  # 011
    vm.registers[3] = 6  # 110
    vm.exec_or("R1", "R2", "R3")
    assert vm.registers[1] == 7


def test_or_increments_pc(vm):
    vm.exec_or("R0", "R1", "R2")
    assert vm.pc == 1


def test_or_big_numbers(vm):
    vm.registers[2] = 8199
    vm.registers[3] = 762
    vm.exec_or("R1", "R2", "R3")
    assert vm.registers[1] == 8959


def test_or_sets_zero_flag(vm):
    vm.registers[2] = 0
    vm.registers[3] = 0
    vm.exec_or("R1", "R2", "R3")
    assert vm.registers[1] == 0
    assert vm.flag_zero
    assert not vm.flag_sign


def test_or_sets_sign_flag(vm):
    vm.registers[2] = to_u16(-1)
    vm.registers[3] = to_u16(-37)
    vm.exec_or("R1", "R2", "R3")
    assert vm.registers[1] == to_u16(-1)
    assert not vm.flag_zero
    assert vm.flag_sign


def test_or_does_not_set_other_flags(vm):
    vm.registers[2] = to_u16(-1)
    vm.registers[3] = to_u16(-1)
    vm.exec_or("R1", "R2", "R3")
    assert vm.registers[1] == to_u16(-1)
    assert not vm.flag_carry
    assert not vm.flag_overflow


def test_or_does_not_clear_other_flags(vm):
    vm.registers[2] = to_u16(-1)
    vm.registers[3] = to_u16(-1)
    vm.flag_carry = True
    vm.flag_overflow = True
    vm.exec_or("R1", "R2", "R3")
    assert vm.flag_carry
    assert vm.flag_overflow


def test_or_does_not_affect_R0(vm):
    vm.registers[1] = 1
    vm.registers[2] = 1
    vm.exec_or("R0", "R1", "R2")
    assert vm.registers[0] == 0


def test_exec_one_delegates_to_xor(vm):
    with patch("hera.vm.VirtualMachine.exec_xor") as mock_exec_xor:
        vm.exec_one(Op("XOR", ["R1", "R2", "R3"]))
        assert mock_exec_xor.call_count == 1
        assert mock_exec_xor.call_args == (("R1", "R2", "R3"), {})


def test_xor_same_numbers(vm):
    vm.registers[2] = 27
    vm.registers[3] = 27
    vm.exec_xor("R1", "R2", "R3")
    assert vm.registers[1] == 0


def test_xor_different_numbers(vm):
    vm.registers[2] = 3  # 011
    vm.registers[3] = 6  # 110
    vm.exec_xor("R1", "R2", "R3")
    assert vm.registers[1] == 5


def test_xor_increments_pc(vm):
    vm.exec_xor("R0", "R1", "R2")
    assert vm.pc == 1


def test_xor_big_numbers(vm):
    vm.registers[2] = 8199
    vm.registers[3] = 762
    vm.exec_xor("R1", "R2", "R3")
    assert vm.registers[1] == 8957


def test_xor_sets_zero_flag(vm):
    vm.registers[2] = 0
    vm.registers[3] = 0
    vm.exec_xor("R1", "R2", "R3")
    assert vm.registers[1] == 0
    assert vm.flag_zero
    assert not vm.flag_sign


def test_xor_sets_sign_flag(vm):
    vm.registers[2] = 0
    vm.registers[3] = to_u16(-37)
    vm.exec_xor("R1", "R2", "R3")
    assert vm.registers[1] == to_u16(-37)
    assert not vm.flag_zero
    assert vm.flag_sign


def test_xor_does_not_set_other_flags(vm):
    vm.registers[2] = 0
    vm.registers[3] = to_u16(-37)
    vm.exec_xor("R1", "R2", "R3")
    assert vm.registers[1] == to_u16(-37)
    assert not vm.flag_carry
    assert not vm.flag_overflow


def test_xor_does_not_clear_other_flags(vm):
    vm.registers[2] = 0
    vm.registers[3] = to_u16(-37)
    vm.flag_carry = True
    vm.flag_overflow = True
    vm.exec_xor("R1", "R2", "R3")
    assert vm.flag_carry
    assert vm.flag_overflow


def test_xor_does_not_affect_R0(vm):
    vm.registers[1] = 1
    vm.registers[2] = 0
    vm.exec_xor("R0", "R1", "R2")
    assert vm.registers[0] == 0
